from flask import current_app
from labbase2.forms import render
from labbase2.forms.filters import strip_input
from labbase2.forms.forms import BaseForm
from wtforms.fields import DateField, IntegerField, SelectField, StringField
from wtforms.validators import DataRequired, Length, NumberRange, Optional

__all__ = ["EditPreparation"]


class EditPreparation(BaseForm):
    """Form to add or edit a plasmid preparation.

    Attributes
    ----------

    """

    preparation_date = DateField(
        label="Date",
        validators=[DataRequired()],
        render_kw=render.custom_field | {"type": "date"},
    )
    method = StringField(
        label="Method",
        validators=[DataRequired(), Length(max=64)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "method"},
    )
    eluent = StringField(
        label="Eluent",
        validators=[DataRequired(), Length(max=32)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Eluent"},
    )
    strain = SelectField(
        label="Strain",
        validators=[DataRequired()],
        choices=[],
        default="DH10B",
        render_kw=render.select_field,
    )
    concentration = IntegerField(
        label="Concentration",
        validators=[DataRequired(), NumberRange(min=1)],
        filters=[lambda x: round(x) if x else x],
        render_kw=render.custom_field | {"type": "number", "min": 1, "step": 1},
    )
    storage_place = StringField(
        label="Location",
        validators=[DataRequired(), Length(max=64)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Location"},
    )
    emptied_date = DateField(
        label="Emptied",
        validators=[Optional()],
        render_kw=render.custom_field | {"type": "date"},
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.strain.choices = [
            (strain, strain) for strain in current_app.config["STRAINS"]
        ]



from labbase2.models import ColumnMapping, Oligonucleotide, db, file
from sqlalchemy import event, func


@event.listens_for(db.session, "deleted_to_detached")
def intercept_deleted_to_detached(session, obj) -> None:
    """Removes the physical file when a File row marked as deleted is
    eventually removed from the database.

    Parameters
    ----------
    session
        The current session.
    obj
        The object that is detached. This can be any Database related object
        but so far this function has only implications for 'File'.


    Returns
    -------
    None
    """

    print("Deleting file.")

    if isinstance(obj, file.BaseFile):
        obj.path.unlink(missing_ok=True)


# TODO: There must be a better option than writing an event for every single child
#  table of BaseEntity.
@event.listens_for(Oligonucleotide, "before_update")
def update_parent(mapper, connection, target) -> None:
    target.timestamp_edited = func.now()


@event.listens_for(ColumnMapping, "before_update")
def update_import_job(mapper, connection, target) -> None:
    target.job.timestamp_edited = func.now()



import json
import secrets
from pathlib import Path
from typing import Optional, Union

from flask import Flask
from labbase2 import logging
from sqlalchemy import select, func


__all__ = ["create_app"]


def create_app(config: Optional[Union[str, Path]] = None, config_dict: Optional[dict] = None, **kwargs) -> Flask:
    """Create an app instance of the labbase2 application.

    Parameters
    ----------
    config : Optional[Union[str, Path]]
        A filename pointing to the configuration file. File has to be in JSON format.
        Filename is supposed to be relative to the instance path.
    config_dict : Optional[dict]
        Additional config parameters for the app. If `config` und `config_dict` contain the same keys, settings from
        `config_dict` will be applied.
    kwargs
        Additional parameters passed to the Flask class during instantiation.
        Supports all parameters of the Flask class except `import_name` and
        `instance_relative_config`, which are hardcoded to `labbase2` and `True`
        respectively.

    Returns
    -------
    Flask
        A configured Flask application instance. If run for the first time,
        an instance folder as well as a sub-folder for uploading files and a SQLite
        database will be created.
    """

    app: Flask = Flask("labbase2", instance_relative_config=True, **kwargs)
    app.config.from_object("labbase2.config.DefaultConfig")

    if config is not None:
        app.config.from_file(config, load=json.load, text=False)
    if config_dict is not None:
        app.config |= config_dict

    # Initialize logging.
    logging.init_app(app)

    # Create upload folder if necessary.
    try:
        Path(app.instance_path, app.config["UPLOAD_FOLDER"]).mkdir(parents=True, exist_ok=True)
    except PermissionError as error:
        app.logger.error("Could not create upload folder due to insufficient permissions!")
        raise error

    # Initiate the database.
    from labbase2.models import db

    db.init_app(app)

    with app.app_context():
        # Create database and add tables (if not yet present).
        db.create_all()

    # Create/update permissions from the config.
    from labbase2.models import Permission, User

    with app.app_context():
        # Add permissions to database.
        for name, description in app.config.get("PERMISSIONS"):
            if (permission := db.session.get(Permission, name)) is None:
                db.session.add(Permission(name=name, description=description))
            else:
                permission.description = description
            db.session.commit()

    from labbase2.models import events

    # If no user with admin rights is in the database, create one.
    with app.app_context():
        first, last, email = app.config.get("USER")

        user_count = db.session.scalar(
            select(func.count())
            .select_from(User)
        )
        admin_count = db.session.scalar(
            select(func.count())
            .select_from(User)
            .where(User.is_active & User.is_admin)
        )

        if user_count == 0:
            app.logger.info("No user in database; create admin as specified in config.")
            admin = User(first_name=first, last_name=last, email=email, is_admin=True)
            admin.set_password("admin")
            admin.permissions = db.session.scalars(select(Permission)).all()
            db.session.add(admin)
        elif admin_count == 0:
            app.logger.info("No active user with admin rights; trying to re-activate admin specified in config.")
            admin = db.session.scalars(select(User).where(User.email == email)).first()
            if admin is not None:
                app.logger.info("Re-activated initial admin '%s' from config.", admin.username)
                admin.is_admin = True
                admin.is_active = True
            else:
                app.logger.info(
                    "Did not find admin user specified in config; create an initial admin from config."
                )
                admin = User(first_name=first, last_name=last, email=email, is_admin=True)
                admin.set_password("admin")
                admin.permissions = db.session.scalars(select(Permission)).all()
                db.session.add(admin)

        app.logger.info("Had %d users and %d admins in database at startup.", user_count, admin_count)

        try:
            db.session.commit()
        except Exception as error:
            app.logger.error("Could not add initial user/admin to database: %s", error)
            raise error

    # Register login_manager with application.
    from labbase2.models.user import login_manager

    login_manager.init_app(app)

    # Register blueprints with application.
    from labbase2 import views

    app.register_blueprint(views.base.bp)
    app.register_blueprint(views.auth.bp)
    app.register_blueprint(views.imports.bp)
    app.register_blueprint(views.chemicals.bp)
    app.register_blueprint(views.comments.bp)
    app.register_blueprint(views.files.bp)
    app.register_blueprint(views.fly_stocks.bp)
    app.register_blueprint(views.requests.bp)
    app.register_blueprint(views.batches.bp)
    app.register_blueprint(views.antibodies.bp)
    app.register_blueprint(views.plasmids.bp)
    app.register_blueprint(views.oligonucleotides.bp)

    # Add custom template filters to Jinja2.
    from labbase2.utils import template_filters

    app.jinja_env.filters["format_date"] = template_filters.format_date
    app.jinja_env.filters["format_datetime"] = template_filters.format_datetime
    app.jinja_env.globals["random_string"] = secrets.token_hex

    return app



from labbase2.models import db
from sqlalchemy import asc, desc

__all__ = ["Filter"]


class Filter:
    """A mixin to add basic filter functionality to a model.

    Notes
    -----
    The mixin is inspired by the template design pattern. The basics method to be
    used is `filter_`. `filter_` uses additional methods to customize the filtering
    process like `_filters` to add filters or `_order_by` to sort the results. The
    mixin provides basic implementations for this. Each model class inheriting from
    this mixin can override one or several of these methods to customize the filter
    process.
    """

    @classmethod
    def filter_(cls, order_by: str, ascending: bool = True, **fields):
        """Create a query object to retrieve matching rows from the database.

        Parameters
        ----------
        order_by : str
            The column name, by which the result shall be ordered.
        ascending : bool
            Order results either ascending or descending. Defaults to `True` (
            ascending).
        fields : **dict
            A list of key-value pairs. The key is a field in the database and the
            value is some constraint which should be applied to this field. See
            `_filters` for details.

        Returns
        -------
        query
            An SQLAlchemy query object. Can be either further customized by adding
            additional query parameters or be used to retrieve matching instances
            from the database.
        """

        query = db.session.query(*cls._entities())

        for join in cls._joins():
            if isinstance(join, tuple):
                query = query.join(*join, isouter=True)
            else:
                query = query.join(join, isouter=True)

        return (
            query.options(*cls._options())
            .filter(*cls._filters(**fields))
            .order_by(*cls._order_by(order_by, ascending))
        )

    @classmethod
    def _filters(cls, **fields) -> list:
        filters = []

        poly_on = cls.__mapper__.polymorphic_on
        if poly_on is not None:
            filters.append(poly_on.is_(cls.__mapper__.polymorphic_identity))

        for field, value in fields.items():
            attr = getattr(cls, field, None)
            if not attr or not value:
                continue

            if value == "none":
                filters.append(attr.is_(None))
            elif value == "any":
                filters.append(attr.isnot(None))
            elif value == "all" or value == 0:
                continue
            # TODO: Improve full-text search (FTS5?)
            elif field == "description":
                value = [f"%{v.strip()}%" for v in value.split()]
                filters += [attr.ilike(v) for v in value]
            elif isinstance(value, int) or field == "entity_type":
                filters.append(attr.is_(value))
            elif isinstance(value, str):
                filters.append(attr.ilike(f"%{value}%"))

        return filters

    @classmethod
    def _order_by(cls, order_by: str, ascending: bool) -> tuple:
        fnc = asc if ascending else desc
        field = getattr(cls, order_by.strip(), cls.id)

        return (fnc(field),)

    @classmethod
    def _options(cls) -> tuple:
        return ()

    @classmethod
    def _entities(cls) -> tuple:
        return (cls,)

    @classmethod
    def _joins(cls) -> tuple:
        return ()

    @classmethod
    def _subquery(cls):
        return None



from flask import has_request_context, request, Flask
from flask_login import current_user as user, AnonymousUserMixin
import logging


__all__ = ["init_app", "RequestFormatter"]


def init_app(app: Flask):
    @app.before_request
    def auto_log_request():
        if not request.path.startswith("/static") and not request.path.startswith("/favicon"):
            app.logger.info("%(method)-6s %(path)s", {"method": request.method, "path": request.url})


class RequestFormatter(logging.Formatter):
    def format(self, record):
        if has_request_context():
            record.url = request.url
            record.remote_addr = request.remote_addr
            record.user = "Anonymous" if isinstance(user, AnonymousUserMixin) else user.username
        else:
            record.url = None
            record.remote_addr = None
            record.user = "Anonymous"

        return super().format(record)



from labbase2 import models
from labbase2.models import db
from sqlalchemy import select, func


def test_first_user_is_admin_and_active(app):
    with app.app_context():
        user = db.session.get(models.User, 1)

        assert user is not None
        assert user.is_admin
        assert user.is_active
        assert user.first_name == "Max"
        assert user.last_name == "Mustermann"
        assert user.email == "test@test.de"


def test_all_permissions_in_db(app):
    with app.app_context():
        permissions_count = db.session.scalar(select(func.count()).select_from(models.Permission))

    assert permissions_count == len(app.config["PERMISSIONS"])


def test_first_user_has_all_permissions(app):
    with app.app_context():
        permissions = db.session.scalars(select(models.Permission)).all()
        user = db.session.get(models.User, 1)

    assert len(permissions) == len(app.config["PERMISSIONS"])
    assert user.permissions == permissions



from typing import Optional, Union

from flask import Blueprint, flash, redirect, request, send_file, url_for, current_app as app
from flask_login import current_user, login_required
from labbase2.models import BaseEntity, BaseFile, EntityFile, db
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required
from werkzeug.utils import secure_filename
from pathlib import Path

from .forms import EditFile, UploadFile

__all__ = ["bp", "upload_file"]


# The blueprint to register all coming routes with.
bp = Blueprint("files", __name__, url_prefix="/files", template_folder="templates")


def upload_file(form, class_, **kwargs) -> Union[BaseFile, EntityFile]:
    file = form.file.data

    if hasattr(form, "filename") and form.filename.data:
        filename = secure_filename(form.filename.data)
    else:
        filename = secure_filename(file.filename)

    db_file = class_(
        user_id=current_user.id,
        filename_exposed=filename,
        note=form.note.data if hasattr(form, "note") else None,
        **kwargs,
    )

    db.session.add(db_file)
    db.session.commit()

    db_file.set_filename()
    db.session.commit()

    try:
        file.save(db_file.path)
    except Exception as error:
        db.session.delete(db_file)
        db.session.commit()
        raise error

    return db_file


@bp.route("/", defaults={"entity_id": None})
@bp.route("/upload/<int:entity_id>", methods=["POST"])
@login_required
@permission_required("Upload files")
def add(entity_id: Optional[int] = None):
    previous_site = request.referrer

    if entity_id and BaseEntity.query.get(entity_id) is None:
        return f"No entity with ID {entity_id}!", 404

    if not (form := UploadFile()).validate():
        return redirect(previous_site)

    if entity_id is not None:
        file = upload_file(form, EntityFile, entity_id=entity_id)
    else:
        file = upload_file(form, BaseFile)

    flash(f"Successfully uploaded '{file.filename_exposed}'!", "success")

    return redirect(previous_site)


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
@permission_required("Upload files")
def edit(id_: int):
    if not (form := EditFile()).validate():
        return str(form.errors)

    if not (file := BaseFile.query.get(id_)):
        return Message.ERROR(f"No file with ID {id_}!")
    elif file.user_id != current_user.id:
        return Message.ERROR("File can only be edited by owner!")
    else:
        file.note = form.note.data
        file.filename_exposed = form.filename.data

    try:
        db.session.commit()
    except Exception as error:
        return str(error)

    return Message.SUCCESS(f"Successfully edited file {file.filename_exposed}!")


@bp.route("/", methods=["GET"], defaults={"id_": None})
@bp.route("/<int:id_>", methods=["GET"])
@login_required
def download(id_: Optional[int] = None):
    if id_ is None:
        return None

    if (file := BaseFile.query.get(id_)) is None:
        return f"No file with ID {id_}!"

    as_attachment = request.args.get("download", False, type=bool)

    return send_file(
        file.path,
        download_name=file.filename_exposed,
        as_attachment=as_attachment,
        mimetype=file.mimetype,
    )


@bp.route("/<int:id_>", methods=["DELETE"])
@login_required
@permission_required("Upload files")
def delete(id_: int):
    if (file := BaseFile.query.get(id_)) is None:
        return Message.ERROR(f"No file with ID {id_}!")

    try:
        db.session.delete(file)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully deleted file {id_}!")


@bp.route("/delete-orphans", methods=["GET"])
@login_required
def delete_orphans():
    if not current_user.is_admin:
        flash("Only admins can delete orphan files!", "danger")
        return redirect(request.referrer)

    # Get all absolute file paths from the database.
    files_db = [file.path for file in BaseFile.query.all()]

    # Get all files from the upload folder.
    dir = Path(app.instance_path, app.config["UPLOAD_FOLDER"])

    deleted = 0

    for file in dir.iterdir():
        if not file.is_file():
            continue

        for i, file_db in enumerate(files_db):
            if file.samefile(file_db):
                files_db.pop(i)
                break
        else:
            app.logger.debug("Delete file %s", file.name)
            deleted += 1
            file.unlink()

    app.logger.info("Deleted %d orphan files.", deleted)
    flash(f"Deleted {deleted} orphan files.", "success")

    return redirect(request.referrer)



import math
import mimetypes
from pathlib import Path

from flask import current_app
from labbase2.models import db
from skimage import io, util
from skimage.transform import resize
from sqlalchemy import func

__all__ = ["BaseFile", "EntityFile"]


class BaseFile(db.Model):
    """Files can be attached to any entity. It is a convenient way to add
    data to entries of the database that do not fit any of the available
    structures. For instance specification sheets for antibodies or
    chemicals, plasmid files, etc...

    Attributes
    ----------
    id : int
        The ID of the filepath for internal identification.
    user_id : int
        The id of the person that has uploaded the filepath.
    filename_internal : str
        The name of the file as it was used to save it to the filesystem.
    filename_exposed : str
        The original filename of the file or the filename provided by the user during
        upload.
    note : str
        A short note about the filepath.
    file_type : str
        The type of the file. This is used for proper mapping and setting up
        inheritance.
    timestamp_uploaded : DateTime
        The time at which the filepath was uploaded. Automatically set by the database.
    timestamp_edited : DateTime
        The time the file entry was last edited.

    Notes
    -----
    There can be different kinds of files (PDFs, images, plasmid files, etc.). Since
    these filepath types sometimes differ in the way they are shown on the website (
    especially images and PDFs) they are modelled through different child classes
    that inherit from this class. Inheritance is implemented as single-table
    inheritance, i.e., all children are represented by a single table/relation in the
    database. They are differentiated by the 'type' attribute.
    """

    __tablename__: str = "base_file"

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    filename_internal = db.Column(db.String(64), nullable=True, unique=True)
    filename_exposed = db.Column(db.String(512), nullable=False)
    note = db.Column(db.String(2048))
    file_type = db.Column(db.String(32), nullable=False)
    timestamp_uploaded = db.Column(
        db.DateTime, nullable=False, server_default=func.now()
    )
    timestamp_edited = db.Column(
        db.DateTime(timezone=True), nullable=True, onupdate=func.now()
    )

    __mapper_args__ = {"polymorphic_on": file_type, "polymorphic_identity": "base_file"}

    @property
    def path(self) -> Path:
        return Path(
            current_app.instance_path,
            current_app.config["UPLOAD_FOLDER"],
            self.filename_internal,
        )

    @property
    def data(self):
        try:
            with open(self.path, mode="r") as file:
                return file.read()
        except UnicodeDecodeError:
            with open(self.path, mode="rb") as file:
                return file.read()

    @property
    def type_(self):
        _, ext = self.filename_internal.rsplit(".", 1)
        match ext.lower():
            case "pdf":
                return "document"
            case "jpg" | "jpeg" | "png" | "tif" | "tiff":
                return "image"
            case "dna" | "gb" | "gbk":
                return "plasmid"
            case _:
                return "misc"

    @property
    def mimetype(self) -> str:
        ext = self.path.suffix.lower()
        return mimetypes.types_map.get(ext, "application/octet-stream")

    def set_filename(self):
        _, ext = self.filename_exposed.split(".", 1)
        self.filename_internal = "{0:07d}.{ext}".format(self.id, ext=ext)

    def resize(self, longest: int):
        if not self.type_ == "image":
            return

        img = io.imread(self.path)
        height, width, _ = img.shape

        if width <= longest and height <= longest:
            return

        if width > height:
            height = math.ceil(height / width * longest)
            width = longest
        else:
            width = math.ceil(width / height * longest)
            height = longest

        resized = resize(img, (height, width), anti_aliasing=True)

        io.imsave(self.path, util.img_as_ubyte(resized))


class EntityFile(BaseFile):
    """A file that is attached to an entity.

    Attributes
    ----------
    id : int
        The internal database ID of this file.
    entity_id : int
        The id of the entity to which this filepath belongs.
    """

    __tablename__ = "entity_file"

    id = db.Column(db.Integer, db.ForeignKey("base_file.id"), primary_key=True)
    entity_id = db.Column(db.Integer, db.ForeignKey("base_entity.id"), nullable=False)

    __mapper_args__ = {"polymorphic_identity": "entity_file"}



from labbase2.forms import render
from labbase2.forms.filters import strip_input
from labbase2.forms.forms import BaseForm
from wtforms.fields import DateField, StringField, TextAreaField
from wtforms.validators import DataRequired, Length, Optional

__all__ = ["EditRequest"]


class EditRequest(BaseForm):
    """Form to add or edit a request.

    Attributes
    ----------

    """

    requested_by = StringField(
        label="Requested by",
        validators=[DataRequired(), Length(max=128)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Requested by"},
    )
    timestamp = DateField(
        label="Date of request",
        validators=[DataRequired()],
        render_kw=render.custom_field | {"type": "date"},
    )
    timestamp_sent = DateField(
        label="Sent",
        validators=[Optional()],
        render_kw=render.custom_field | {"type": "date"},
    )
    note = TextAreaField(
        label="Note",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"rows": 4},
    )



from flask import Blueprint
from flask_login import current_user, login_required
from labbase2.models import Comment, db
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required

from .forms import EditComment

__all__ = ["bp"]


bp = Blueprint("comments", __name__, url_prefix="/comment", template_folder="templates")


@bp.route("/attach/<int:entity_id>", methods=["POST"])
@login_required
@permission_required("Write comment")
def add(entity_id: int):
    form = EditComment()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    comment = Comment(entity_id=entity_id, user_id=current_user.id)
    form.populate_obj(comment)

    try:
        db.session.add(comment)
        db.session.commit()
    except Exception as error:
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully added comment to '{comment.entity.label}'!")


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
@permission_required("Write comment")
def edit(id_: int):
    form = EditComment()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if (comment := Comment.query.get(id_)) is None:
        return Message.ERROR(f"No comment found with ID {id_}!")

    if comment.user_id != current_user.id:
        return Message.ERROR("Comment can only be edited by original author!")

    form.populate_obj(comment)

    try:
        db.session.commit()
    except Exception as error:
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully edited comment {id_}!")


@bp.route("/<int:id_>", methods=["DELETE"])
@login_required
@permission_required("Write comment")
def delete(id_):
    if not (comment := Comment.query.get(id_)):
        return Message.ERROR(f"No comment with ID {id_}!")

    if comment.user_id != current_user.id:
        return Message.ERROR("Comment can only be deleted by original author!")

    try:
        db.session.delete(comment)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully deleted comment {id_}!")



from sqlite3 import IntegrityError

from flask import Blueprint
from flask import current_app as app
from flask import flash, render_template, request
from flask_login import current_user, login_required
from labbase2.models import Chemical, db
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required
from labbase2.views.batches.forms import EditBatch
from labbase2.views.comments.forms import EditComment
from labbase2.views.files.forms import UploadFile

from . import stock_solutions
from .forms import EditChemical, FilterChemical
from .stock_solutions.forms import EditStockSolution

__all__ = ["bp"]


# The blueprint to register all coming blueprints with.
bp = Blueprint(
    "chemicals", __name__, url_prefix="/chemical", template_folder="templates"
)

bp.register_blueprint(stock_solutions.bp)


@bp.route("/", methods=["GET"])
@login_required
def index():
    page = request.args.get("page", 1, type=int)
    form = FilterChemical(request.args)

    data = form.data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = Chemical.filter_(**data)
    except Exception as err:
        flash(str(err), "danger")
        entities = Chemical.filter_(order_by="label")

    return render_template(
        "chemicals/main.html",
        filter_form=form,
        import_file_form=UploadFile(),
        add_form=EditChemical(formdata=None),
        entities=entities.paginate(page=page, per_page=app.config["PER_PAGE"]),
        total=Chemical.query.count(),
        title="Chemicals",
    )


@bp.route("/", methods=["POST"])
@login_required
@permission_required("Add chemicals")
def add():
    form = EditChemical()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    chemical = Chemical()
    chemical.origin = f"Created via import form by {current_user.username}."
    form.populate_obj(chemical)

    try:
        db.session.add(chemical)
        db.session.commit()
    except IntegrityError as error:
        return Message.ERROR(error)
    except Exception as error:
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully added chemical '{chemical.label}'!")


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
@permission_required("Add chemicals")
def edit(id_: int):
    form = EditChemical()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if not (chemical := Chemical.query.get(id_)):
        return Message.ERROR(f"No chemical with ID {id_}!")

    form.populate_obj(chemical)

    try:
        db.session.commit()
    except Exception as error:
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully edited chemical '{chemical.label}'!")


@bp.route("/<int:id_>", methods=["DELETE"])
@login_required
@permission_required("Add chemicals")
def delete(id_: int):
    if (chemical := Chemical.query.get(id_)) is None:
        return Message.ERROR(f"No chemical with ID {id_}!")

    try:
        db.session.delete(chemical)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully deleted chemical '{chemical.label}'!")


@bp.route("/<int:id_>", methods=["GET"])
@login_required
def details(id_: int):
    if (chemical := Chemical.query.get(id_)) is None:
        return Message.ERROR(f"No chemical found with ID {id_}!")

    return render_template(
        "chemicals/details.html",
        chemical=chemical,
        form=EditChemical(formdata=None, obj=chemical),
        comment_form=EditComment,
        file_form=UploadFile,
        batch_form=EditBatch,
        stock_form=EditStockSolution,
    )


@bp.route("/export/<string:format_>/", methods=["GET"])
@login_required
@permission_required("Export content")
def export(format_: str):
    data = FilterChemical(request.args).data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = Chemical.filter_(**data)
    except Exception:
        return "An internal error occured! Please inform the admin!", 500

    match format_:
        case "csv":
            return Chemical.export_to_csv(entities)
        case "json":
            return Chemical.export_to_json(entities)
        case _:
            return Message.ERROR(f"Unsupported format: {format_}"), 400



import pytest
from labbase2 import create_app


@pytest.fixture
def app():
    app = create_app(
        config_dict={
            "TESTING": True,
            "WTF_CSRF_ENABLED": False,
            "SERVER_NAME": "localhost",
            "SQLALCHEMY_DATABASE_URI": "sqlite:///:memory:",
            "USER": ["Max", "Mustermann", "test@test.de"],
        }
    )

    yield app


@pytest.fixture
def client(app):
    return app.test_client()



from flask_wtf import FlaskForm
from labbase2.forms import render
from wtforms.fields import FieldList, SelectField, SubmitField

__all__ = ["MappingForm"]


class MappingForm(FlaskForm):
    """A form to map columns from uploaded files to the fields of an entity.

    Parameters
    ----------
    fields : list[str]
        A list of names to be shown as labels in the forms. This represents the
        importable fields
        of the respective entity (e.g. plasmid). However, the entity fields are not
        matched
        against this list. Thus, the names can be changed to increase readability. By
        default,
        dashes and underscores are replace by spaces and the names are capitalized.
    choices : list[tuple[str, str]]
        A list of choices to choose from for the mapping. This should be the column
        names of the
        uploaded file.

    Notes
    -----
    The class should be instantiated with the default values for each field.

    >>> defaults = ["name", "-", "-"]
    >>> MappingForm(data={"mapping": defaults})

    The defaults are usually taken from the database and represent the saved state of
    the user from his last edit of this import. The order of the default has to match
    the order of the `fields` parameter.

    IMPORTANT: The mapping field is a `FieldList` of `SelectField`s. In order to allow
    instantiating the class with default values with dynamic choices (they differ for
    each uploaded file), the `SelectField`s do not validate the provided default
    values. Thus, extra care must be taken to ensure integrity of the instance.
    """

    mapping = FieldList(
        SelectField(
            "Name",
            choices=[(None, "-")],
            default=None,
            validate_choice=False,
            render_kw={"class": "form-select form-select-sm"},
        )
    )
    submit = SubmitField(label="Update mapping", render_kw=render.submit_field)

    def __init__(
        self, fields: list[str], choices: list[tuple[str, str]], *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)

        if len(fields) != len(self.mapping):
            raise ValueError("'fields' must have the same length as mapping")

        for field, mapping in zip(fields, self.mapping):
            # Increase readability of the label.
            label = field.replace("_", " ")
            label = label.replace("-", " ")

            mapping.choices += choices
            mapping.label = label.capitalize()



from flask import Blueprint
from flask import current_app as app
from flask import flash, render_template, request
from flask_login import current_user, login_required
from labbase2.models import StockSolution, db
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required
from labbase2.views.chemicals.forms import EditChemical
from labbase2.views.files.forms import UploadFile

from .forms import EditStockSolution, FilterStockSolution

__all__ = ["bp"]


# The blueprint to register all coming blueprints with.
bp = Blueprint(
    "solutions", __name__, url_prefix="/stock-solutions", template_folder="templates"
)


@bp.route("/", methods=["GET"])
@login_required
def index():
    page = request.args.get("page", 1, type=int)
    form = FilterStockSolution(request.args)

    data = form.data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = StockSolution.filter_(**data)
    except Exception as err:
        flash(str(err), "danger")
        entities = StockSolution.filter_(order_by="label")

    return render_template(
        "stock_solutions/main.html",
        filter_form=form,
        add_form=EditStockSolution(formdata=None),
        entities=entities.paginate(page=page, per_page=app.config["PER_PAGE"]),
        total=StockSolution.query.count(),
        title="Stock Solutions",
    )


@bp.route("/<int:chemical_id>", methods=["POST"])
@login_required
@permission_required("Add stock solution")
def add(chemical_id: int):
    form = EditStockSolution()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    solution = StockSolution(chemical_id=chemical_id, user_id=current_user.id)
    form.populate_obj(solution)

    try:
        db.session.add(solution)
        db.session.commit()
    except Exception as error:
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully added stock of '{solution.chemical.label}'!")


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
@permission_required("Add stock solution")
def edit(id_: int):
    form = EditStockSolution()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if (solution := StockSolution.query.get(id_)) is None:
        return Message.ERROR(f"No stock solution with ID {id_}!")

    form.populate_obj(solution)

    try:
        db.session.commit()
    except Exception as error:
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully edited stock solution {id_}!")


@bp.route("/<int:id_>", methods=["DELETE"])
@login_required
@permission_required("Add stock solution")
def delete(id_):
    if (solution := StockSolution.query.get(id_)) is None:
        return Message.ERROR(f"No stock solution with ID {id_}!")

    try:
        db.session.delete(solution)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return Message.SUCCESS(f"Successfully deleted stock solution {id_}!")


@bp.route("/<int:id_>/<string:format_>", methods=["GET"])
@login_required
def details(id_: int, format_: str):
    if (solution := StockSolution.query.get(id_)) is None:
        return Message.ERROR(f"No stock solution wth ID {id_}!")

    match format_:
        case "long":
            template = "stock_solutions/details.html"
        case "tab":
            template = "stock_solutions/details-tab.html"
        case _:
            return Message.ERROR(f"Invalid format '{format_}'!")

    edit_form = EditStockSolution(None, obj=solution)

    return render_template(
        template,
        stock=solution,
        form=edit_form,
        file_form=UploadFile,
        chemical_form=EditChemical(None, obj=solution.chemical),
    )



import pandas as pd
from sqlalchemy import inspect

__all__ = ["Importer"]


class Importer:

    import_attr: tuple = (("id", "ID"),)
    not_updatable: tuple = ("id",)

    def update(self, **kwargs) -> None:
        """Update attributes of an entity.

        Parameters
        ----------
        kwargs
            Attributes to be updated and the new value.

        Returns
        -------
        None

        Notes
        -----
        The changes are not committed automatically to the database, i.e.,
        any changes done via update will not be saved if the calling code
        does not commit these changes on its own.
        """

        for attr, value in kwargs.items():
            if hasattr(self, attr):
                setattr(self, attr, value)

    @classmethod
    def importable_fields(cls) -> list:
        fields = []
        for column in inspect(cls).columns:
            if column.info.get("importable", False):
                fields.append(column.name)
        return fields

    @classmethod
    def from_record(cls, rec: dict, update: bool = False):

        rec = cls.process_record(rec=rec)

        if update:
            if id_ := rec.pop("id", None):
                entity = cls.query.get(id_)
            elif label := rec.pop("label", None):
                entity = cls.query.filter(cls.label == label).first()
            else:
                entity = None

            if not entity:
                return

            for key in rec:
                if key in cls.not_updatable:
                    del rec[key]

            entity.update(**rec)

        else:
            if "id" in rec:
                del rec["id"]

            entity = cls(**rec)

        return entity

    @classmethod
    def process_record(cls, rec: dict) -> dict:
        return {k: v for k, v in rec.items() if not pd.isnull(v)}

    # @classmethod
    # def import_form(cls, columns: list, *args, **kwargs) -> ImportEntity:
    #     data = {'mappings': len(columns) * [[]]}
    #
    #     form = ImportEntity(clss=cls.__name__, data=data, *args, **kwargs)
    #
    #     for column, field in zip(columns, form.mappings):
    #         field.label = column
    #         field.choices += cls.import_attr
    #
    #     return form

    @staticmethod
    def process_formdata(data: dict) -> dict:
        return {k: v for k, v in data.items() if v}



from Bio.Seq import Seq
from flask import Blueprint
from flask import current_app as app
from flask import flash, render_template, request
from flask_login import current_user, login_required
from labbase2.models import Oligonucleotide, db
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required
from labbase2.views.comments.forms import EditComment
from labbase2.views.files.forms import UploadFile
from sqlalchemy import func
from sqlalchemy.exc import IntegrityError

from .forms import EditOligonucleotide, FilterOligonucleotide, FindOligonucleotide
from .lcsfinder import LCSFinder

__all__ = ["bp"]


# The blueprint to register all coming blueprints with.
bp = Blueprint(
    "oligonucleotides",
    __name__,
    url_prefix="/oligonucleotide",
    template_folder="templates",
)


@bp.route("/", methods=["GET"])
@login_required
def index():
    page = request.args.get("page", 1, type=int)
    form = FilterOligonucleotide(request.args)

    data = form.data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = Oligonucleotide.filter_(**data)
    except Exception as error:
        flash(str(error), "danger")
        entities = Oligonucleotide.filter_(order_by="label")

    return render_template(
        "oligonucleotides/main.html",
        filter_form=form,
        import_file_form=UploadFile(),
        add_form=EditOligonucleotide(formdata=None),
        entities=entities.paginate(page=page, per_page=app.config["PER_PAGE"]),
        total=Oligonucleotide.query.count(),
        title="Oligonucleotides",
    )


@bp.route("/<int:id_>", methods=["GET"])
@login_required
def details(id_: int):
    if (oligonucleotide := Oligonucleotide.query.get(id_)) is None:
        return Message.ERROR(f"No oligonucleotide found with ID {id_}!")

    return render_template(
        "oligonucleotides/details.html",
        oligonucleotide=oligonucleotide,
        form=EditOligonucleotide(None, obj=oligonucleotide),
        comment_form=EditComment,
        file_form=UploadFile,
    )


@bp.route("/", methods=["POST"])
@login_required
@permission_required("Add oligonucleotide")
def add():
    form = EditOligonucleotide()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    oligonucleotide = Oligonucleotide()
    oligonucleotide.origin = f"Created via import form by {current_user.username}."
    form.populate_obj(oligonucleotide)

    try:
        db.session.add(oligonucleotide)
        db.session.commit()
    except IntegrityError as error:
        return Message.ERROR(error)
    except Exception as error:
        return Message.ERROR(error)
    else:
        return Message.SUCCESS(
            f"Successfully added oligonucleotide '{oligonucleotide.label}'!"
        )


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
@permission_required("Add oligonucleotide")
def edit(id_: int):
    form = EditOligonucleotide()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if not (oligonucleotide := Oligonucleotide.query.get(id_)):
        return Message.ERROR(f"No oligonucleotide with ID {id_}!")

    if oligonucleotide.owner_id != current_user.id:
        return Message.ERROR("Oligonucleotide can only be edited by owner!")

    form.populate_obj(oligonucleotide)

    try:
        db.session.commit()
    except Exception as error:
        return Message.ERROR(error)

    return Message.SUCCESS(
        f"Successfully edited oligonucleotide {oligonucleotide.label}!"
    )


@bp.route("/<int:id_>", methods=["DELETE"])
@login_required
@permission_required("Add oligonucleotide")
def delete(id_):
    if (oligonucleotide := Oligonucleotide.query.get(id_)) is None:
        return Message.ERROR(f"No oligonucleotide with ID {id_}!")

    if oligonucleotide.owner_id != current_user.id:
        return Message.ERROR("Only owner can delete oligonucleotide!")

    try:
        db.session.delete(oligonucleotide)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return Message.SUCCESS(
        f"Successfully deleted oligonucleotide '{oligonucleotide.label}'!"
    )


@bp.route("/find", methods=["GET", "POST"])
@login_required
def find():
    form = FindOligonucleotide()
    results = []

    if form.validate_on_submit():
        seq = form.sequence.data
        min_match = form.min_match.data
        max_len = form.max_len.data
        length = len(seq)

        if form.reverse_complement.data:
            seq = str(Seq(seq).reverse_complement())

        lcsfinder = LCSFinder(seq)
        oligonucleotides = Oligonucleotide.query.filter(
            func.length(Oligonucleotide.sequence) >= min_match,
            func.length(Oligonucleotide.sequence) <= max_len,
        )

        for oligonucleotide in oligonucleotides:
            try:
                lcsresult = lcsfinder(oligonucleotide.sequence)
            except Exception as error:
                flash(str(error), "danger")
                continue

            if lcsresult.length >= min_match:
                results.append((oligonucleotide, lcsresult))

        # Sort results from longest to shorted common substring.
        results.sort(key=lambda x: x[1].length, reverse=True)
    else:
        length = 0

    return render_template(
        "oligonucleotides/find.html",
        filter_form=form,
        entities=results,
        length=length,
        title="Find oligonucleotides",
    )


@bp.route("/export/<string:format_>/", methods=["GET"])
@login_required
@permission_required("Export content")
def export(format_: str):
    data = FilterOligonucleotide(request.args).data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = Oligonucleotide.filter_(**data)
    except Exception as error:
        return Message.ERROR(error)

    match format_.lower():
        case "csv":
            return Oligonucleotide.export_to_csv(entities)
        case "json":
            return Oligonucleotide.export_to_json(entities)
        case "fasta":
            return Oligonucleotide.to_fasta(entities)
        case _:
            return Message.ERROR(f"Unsupported format '{format_}'!")



from labbase2.models import db
from sqlalchemy import func

__all__ = ["ImportJob", "ColumnMapping"]


class ImportJob(db.Model):

    __tablename__ = "import_job"

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    timestamp = db.Column(db.DateTime, default=func.now())
    timestamp_edited = db.Column(db.DateTime, default=func.now())
    file_id = db.Column(db.Text, db.ForeignKey("base_file.id"), nullable=False)
    is_finished = db.Column(db.Boolean, default=False, nullable=False)
    entity_type = db.Column(db.Text, nullable=False)

    # One-to-many relationships.
    mappings = db.relationship(
        "ColumnMapping", backref="job", cascade="all, delete-orphan", lazy=True
    )

    file = db.relationship(
        "BaseFile",
        backref="import_job",
        lazy=True,
        cascade="all, delete",
        single_parent=True,
    )

    def get_file(self):
        pass


class ColumnMapping(db.Model):

    __tablename__ = "column_mapping"

    job_id = db.Column(db.Integer, db.ForeignKey("import_job.id"), primary_key=True)
    mapped_field = db.Column(db.String, primary_key=True)
    input_column = db.Column(db.String)



from flask_login import current_user
from labbase2.forms import render
from labbase2.forms.filters import strip_input
from labbase2.forms.forms import EditEntityForm, FilterForm
from labbase2.forms.validators import ContainsNot
from labbase2.models import User
from wtforms.fields import DateField, SelectField, StringField, TextAreaField
from wtforms.validators import DataRequired, Length, Optional

__all__ = ["FilterFlyStocks", "EditFlyStock"]


class FilterFlyStocks(FilterForm):
    """ """

    label = StringField(
        label="Label",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Label"},
        description="The label, i.e. the name of the entity.",
    )
    owner_id = SelectField(
        label="Owner",
        validators=[Optional()],
        choices=[(0, "All")],
        default=lambda: current_user.id,
        coerce=int,
        render_kw=render.select_field,
        description="The owner of the fly stock.",
    )
    short_genotype = StringField(
        label="Short Genotype",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "m6[3xcs]"},
        description="The genotype of the fly stock.",
    )
    chromosome_x = StringField(
        label="Chromosome X", validators=[Optional()], render_kw=render.custom_field
    )
    chromosome_y = StringField(
        label="Chromosome Y", validators=[Optional()], render_kw=render.custom_field
    )
    chromosome_2 = StringField(
        label="Chromosome 2", validators=[Optional()], render_kw=render.custom_field
    )
    chromosome_3 = StringField(
        label="Chromosome 3", validators=[Optional()], render_kw=render.custom_field
    )
    chromosome_4 = StringField(
        label="Chromosome 4", validators=[Optional()], render_kw=render.custom_field
    )
    source = StringField(
        label="Source",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "anne uv"},
        description="Where the stock was retrieved from.",
    )
    reference = StringField(
        label="Reference",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field
        | {"id": "filter-form-reference", "placeholder": "byri"},
        description="A paper that was added as a reference for this stock.",
    )
    discarded = SelectField(
        label="Discarded",
        choices=[
            ("all", "All"),
            ("recent", "Available only"),
            ("discarded", "Discarded only"),
        ],
        render_kw=render.select_field,
        description="""
        Shall only be discarded or non-discarded (recent) stocks be returned?
        """,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.order_by.choices += [
            ("label", "Label"),
            ("owner_id", "Owner"),
            ("short_genotype", "Short Genotype"),
            ("source", "Source"),
            ("discarded", "Discarded"),
        ]
        users = (
            User.query.with_entities(User.id, User.username)
            .order_by(User.username)
            .all()
        )
        self.owner_id.choices += users

    def fields(self) -> list:
        return [
            self.label,
            self.owner_id,
            self.short_genotype,
            self.chromosome_x,
            self.chromosome_y,
            self.chromosome_2,
            self.chromosome_3,
            self.chromosome_4,
            self.source,
            self.reference,
            self.discarded,
            self.order_by,
            self.ascending,
        ]


class EditFlyStock(EditEntityForm):
    """Form to add or edit a general fly stock.

    Attributes
    ----------

    """

    owner_id = SelectField(
        "Owner",
        choices=[(0, "-")],
        coerce=int,
        render_kw=render.select_field,
    )
    short_genotype = StringField(
        "Short Genotype",
        validators=[Optional(), Length(max=2048)],
        render_kw=render.custom_field,
    )
    chromosome_xa = StringField(
        "Chromosome XA",
        validators=[DataRequired(), ContainsNot(["/"]), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field,
        default="+",
    )
    chromosome_xb = StringField(
        "Chromosome XB",
        validators=[DataRequired(), ContainsNot(["/"]), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field,
        default="+",
    )
    chromosome_y = StringField(
        "Chromosome Y",
        validators=[DataRequired(), ContainsNot(["/"]), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field,
        default="+",
    )
    chromosome_2a = StringField(
        "Chromosome 2A",
        validators=[DataRequired(), ContainsNot(["/"]), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field,
        default="+",
    )
    chromosome_2b = StringField(
        "Chromosome 2B",
        validators=[DataRequired(), ContainsNot(["/"]), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field,
        default="+",
    )
    chromosome_3a = StringField(
        "Chromosome XA",
        validators=[DataRequired(), ContainsNot(["/"]), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field,
        default="+",
    )
    chromosome_3b = StringField(
        "Chromosome 3B",
        validators=[DataRequired(), ContainsNot(["/"]), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field,
        default="+",
    )
    chromosome_4a = StringField(
        "Chromosome 4A",
        validators=[DataRequired(), ContainsNot(["/"]), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field,
        default="+",
    )
    chromosome_4b = StringField(
        "Chromosome 4B",
        validators=[DataRequired(), ContainsNot(["/"]), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field,
        default="+",
    )
    created_date = DateField(
        "Created",
        validators=[Optional()],
        render_kw=render.custom_field | {"type": "date"},
    )
    location = StringField(
        "Location",
        validators=[Optional(), Length(max=64)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Location"},
    )
    source = StringField(
        "Source",
        validators=[Optional(), Length(max=512)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Source"},
    )
    documentation = TextAreaField(
        "Documentation",
        validators=[Optional(), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Documentation"},
    )
    reference = StringField(
        "Reference",
        validators=[Optional(), Length(max=512)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Reference"},
    )
    discarded_date = DateField(
        "Discarded",
        validators=[Optional()],
        render_kw=render.custom_field | {"type": "date"},
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        users = (
            User.query.with_entities(User.id, User.username)
            .order_by(User.username)
            .all()
        )
        self.owner_id.choices += users



import secrets

from flask_login import LoginManager, UserMixin
from labbase2.models import db
from labbase2.models.mixins import Export
from sqlalchemy import func
from sqlalchemy.ext.hybrid import hybrid_property
from werkzeug.security import check_password_hash, generate_password_hash

__all__ = ["login_manager", "User", "Permission", "user_permissions", "ResetPassword"]


login_manager = LoginManager()
login_manager.login_view = "auth.login"
login_manager.login_message_category = "warning"

# Map users to their roles in a many-to-many relationship.
user_permissions = db.Table(
    "user_permissions",
    db.Column("user_id", db.Integer, db.ForeignKey("user.id"), primary_key=True),
    db.Column("role_id", db.Integer, db.ForeignKey("permission.name"), primary_key=True),
)


class User(db.Model, UserMixin, Export):
    """A person (user) that should have access to the database.

    Attributes
    ----------
    id : int
        An internal identifier of the person.
    first_name : str
        The person's given name.
    last_name : str
        The person's last name.
    email : str
        The email address.
    password_hash : str
        This is a hash of the password set by the user. This attribute is not set
        manually. Instead, the clear password is passed to the 'set_password' method.
    file_picture_id : int
        ID of a file serving as the profile picture.
    timestamp_created : Datetime
        The time the profile was created. Automatically set by the database.
    timestamp_last_login : Datetime
        Time of last log in.
    timezone : str
        Preferred timezone to show all times in.
    is_active : bool
        A logical flag indicating if the user is active. If `False` the user can no
        longer sign in to the application.
    is_admin : bool
        A logical flag indicating if the user is an admin. Admins can do everything
        in the application.
    picture : File
        The `File` instance for the profile picture.
    comments : list[Comment]
        A list of all comments that were created by this person.
    plasmids : list[Plasmid]
        A list of all plasmids owned by this user.
    glycerol_stocks : list[GlycerolStock]
        A list of all glycerol stocks owned by this user.
    oligonucleotides : list[Oligonucleotide]
        A list of all oligonucleotides owned by this user. NOTE: This was previously
        called 'primers' but oligonucleotide is more accurate.
    preparations : list[Preparation]
        All plasmid preparations done by this user. This can be different than the
        owner of the plasmid to which the preparation belongs.
    dilutions : list[Dilution]
        A list of antibody dilutions determined by this person.
    files : list[BaseFile]
        All files uploaded by this person.
    modifications : list[Modification]
        All modifications of fly stocks done by this user. This can be different from
        the owner of the fly stock.
    fly_stocks : list[FlyStock]
        A list of all fly stocks owned by this user.
    responsibilities : list[Chemical]
        All chemicals for which this user is responsible.
    stock_solutions : list[StockSolution]
        All stock solutions prepared by this user.
    import_jobs : list[ImportJob]
        A list of all import jobs of this user.
    resets : list[ResetPassword]
        A list of reset requests for the password made by this person. The database
        scheme theoretically allows to have several such resets active for one user
        but the application should make sure to delete any previous request for a new
        password once another request is started.
    permissions : list[Permission]
        Roles of this user. The set of roles determines hat a user can see and do in
        the application.


    Notes
    -----
    A person is ultimately identified by its id, which is used internally for
    everything. However, for convenience a user can login to the app using his/her
    email address. Therefore, the email address has to be unique among all users. The
    same applies to the username.

    Users inherit from the UserMixin class of the flask_login module. The roles
    system is self-implemented in the app.utils module via a simple decorator.
    """

    __tablename__: str = "user"

    id = db.Column(db.Integer, primary_key=True)
    first_name = db.Column(db.String(64), nullable=False)
    last_name = db.Column(db.String(64), nullable=False)
    email = db.Column(db.String(128), unique=True, nullable=False)
    password_hash = db.Column(db.String(512), nullable=False)
    file_picture_id = db.Column(db.Integer, db.ForeignKey("base_file.id"), nullable=True)
    timestamp_created = db.Column(db.DateTime, server_default=func.now(), nullable=False)
    timestamp_last_login = db.Column(db.DateTime, nullable=True)
    timezone = db.Column(db.String(64), nullable=False, default="UTC")
    is_active = db.Column(db.Boolean, nullable=False, default=True)
    is_admin = db.Column(db.Boolean, nullable=False, default=False)

    # Set relationship for profile picture.
    picture = db.relationship(
        "BaseFile",
        backref="profile",
        lazy=True,
        foreign_keys=[file_picture_id],
        cascade="all, delete-orphan",
        single_parent=True,
    )

    # One-to-many relationships.
    comments = db.relationship(
        "Comment",
        backref="user",
        order_by="Comment.timestamp_created.desc()",
        lazy=True,
    )
    plasmids = db.relationship(
        "Plasmid",
        backref="owner",
        lazy=True,
        order_by="Plasmid.timestamp_created.desc()",
        foreign_keys="Plasmid.owner_id",
    )
    glycerol_stocks = db.relationship("GlycerolStock", backref="owner", lazy=True)
    oligonucleotides = db.relationship(
        "Oligonucleotide",
        backref="owner",
        lazy=True,
        order_by="Oligonucleotide.timestamp_created.desc()",
        foreign_keys="Oligonucleotide.owner_id",
    )
    preparations = db.relationship("Preparation", backref="owner", lazy=True)
    dilutions = db.relationship("Dilution", backref="user", lazy=True)
    files = db.relationship(
        "EntityFile", backref="user", lazy=True, foreign_keys="EntityFile.user_id"
    )
    modifications = db.relationship("Modification", backref="user", lazy=True)
    fly_stocks = db.relationship(
        "FlyStock", backref="owner", lazy=True, foreign_keys="FlyStock.owner_id"
    )
    responsibilities = db.relationship(
        "Chemical",
        backref="responsible",
        lazy=True,
        foreign_keys="Chemical.owner_id",
    )
    stock_solutions = db.relationship("StockSolution", backref="owner", lazy=True)
    import_jobs = db.relationship(
        "ImportJob", backref="user", lazy=True, order_by="ImportJob.timestamp.asc()"
    )
    resets = db.relationship(
        "ResetPassword", backref="user", lazy=True, cascade="all, delete-orphan"
    )

    # Many-to-many relationships.
    permissions = db.relationship(
        "Permission",
        secondary=user_permissions,
        lazy="subquery",
        backref=db.backref("users", lazy=True),
    )

    @hybrid_property
    def username(self):
        return self.first_name + " " + self.last_name

    @username.expression
    def username(cls):
        return cls.first_name + " " + cls.last_name

    @property
    def form_permissions(self) -> dict:
        out = {}
        for permission in self.permissions:
            out[permission.name.lower().replace(" ", "_")] = True

        return out

    def set_password(self, password: str) -> None:
        """Creates a hash that is stored in the database to validate the user's
        password.

        Parameters
        ----------
        password : str
            A string from which the hash shall be created.

        Returns
        -------
        None

        Notes
        -----
        Of course, passwords are not stored as clear text in the database. Instead,
        a hash is generated from the
        password the user enters and that hash is stored. It is not possible to
        reconstruct the original password but
        the same hash is generated from the same password. Thus, the hash can be used
        to verify users.

        Currently, this method accepts all strings. In the future users will maybe be
        forced to add passwords that
        comply with certain restrictions to ensure a certain level of security.
        """

        self.password_hash = generate_password_hash(password)

    def verify_password(self, password: str) -> bool:
        """Creates the hash from the 'password' parameter and checks this
        hash against the hash deployed in the database.

        Parameters
        ----------
        password : str
            The password to be checked.

        Returns
        -------
        bool
            Returns True if the hash is the same as in the database and False
            otherwise.

        """

        return check_password_hash(self.password_hash, password)

    def has_permission(self, permission: str) -> bool:
        permission_db = db.session.get(Permission, permission)
        if permission_db is None:
            raise ValueError(f"Unknown permission '{permission}'!")

        return permission_db in self.permissions

    @classmethod
    def generate_password(cls) -> str:
        return secrets.token_hex(6)


class Permission(db.Model, Export):
    """Roles a user could possibly have.

    Attributes
    ----------
    name : str
        A descriptive name of the role.
    description : str
        A more accurate description of what this permission allows a user to do.

    Notes
    -----
    On the database levels roles don't have any meaning. They are just names. Meaning
    is conferred by the application.
    """

    __tablename__: str = "permission"

    name = db.Column(db.String(32), primary_key=True, nullable=False)
    description = db.Column(db.String(512), nullable=True)


class ResetPassword(db.Model):
    """Requests to reset the password of a user.

    id : int
        The database ID of the request.
    user_id : int
        Database ID of the user for which the password should be reset.
    key : str
        A long random key used to log in a user once.
    timeout : datetime
        The datetime at which this requests becomes invalid. The user then can no
        longer use the key to log in and has to start a new request.
    """

    __tablename__: str = "reset_password"

    token = db.Column(db.String(64), primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False, unique=True)
    timeout = db.Column(db.DateTime, nullable=False)


@login_manager.user_loader
def _load_user(id_: str) -> User | None:
    """Load a user from database by ID.

    Parameters
    ----------
    id_ : str
        The internal database ID of the user. The database ID of a user is an
        integer. However, the user ID for the session is stored in unicode.
        So the value has to be converted to integer before.

    Returns
    -------
    User | None
        Either the user if a valid existing ID was provided or `None` otherwise.
    """

    try:
        id_ = int(id_)
    except ValueError:
        print("Invalid ID provided: {}".format(id))
        return None
    else:
        return User.query.get(id_)



from datetime import date, datetime
from typing import Optional
from zoneinfo import ZoneInfo

from flask_login import current_user

__all__ = ["format_date", "format_datetime"]


def format_date(x: Optional[date] = None) -> str:
    if not x:
        return ""

    return x.strftime("%B %d, %Y")


def format_datetime(x: Optional[datetime] = None) -> str:
    if not x:
        return ""

    tz = getattr(current_user, "timezone", "Europe/Berlin")

    return x.astimezone(ZoneInfo(tz)).strftime("%B %d, %Y %I:%M %p")



from flask import Blueprint
from flask_login import login_required
from labbase2.forms.utils import errors2messages
from labbase2.models import Request, db

from .forms import EditRequest

__all__ = ["bp"]


bp = Blueprint("requests", __name__, url_prefix="/request", template_folder="templates")


@bp.route("/<int:entity_id>", methods=["POST"])
@login_required
def add(entity_id: int):
    if (form := EditRequest()).validate():
        request = Request(entity_id=entity_id)
        form.populate_obj(request)

        try:
            db.session.add(request)
            db.session.commit()
        except Exception as err:
            return str(err), 400
        else:
            return f"Successfully added request!", 201

    else:
        print(form.errors)
        return errors2messages(form.errors), 400


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
def edit(id_: int):
    if (form := EditRequest()).validate():
        if not (request := Request.query.get(id_)):
            return f"No request with ID {id_}!", 404
        else:
            form.populate_obj(request)

        try:
            db.session.commit()
        except Exception as err:
            return str(err), 400
        else:
            return f"Successfully edited request {id_}!", 200
    else:
        return errors2messages(form.errors), 400


@bp.route("/<int:id_>", methods=["DELETE"])
@login_required
def delete(id_):
    if not (request := Request.query.get(id_)):
        return f"No comment with ID {id_}!", 404
    else:
        try:
            db.session.delete(request)
            db.session.commit()
        except Exception as err:
            db.session.rollback()
            return str(err), 400
        else:
            return f"Successfully deleted request {id_}!", 200



from labbase2.forms import render
from labbase2.forms.filters import strip_input
from labbase2.forms.forms import BaseForm, FilterForm
from labbase2.models import Batch
from wtforms.fields import (
    DateField,
    DecimalField,
    IntegerField,
    SelectField,
    StringField,
)
from wtforms.validators import DataRequired, Length, NumberRange, Optional

__all__ = ["FilterBatch", "EditBatch"]


class FilterBatch(FilterForm):
    id = IntegerField(
        "ID",
        validators=[Optional()],
        render_kw=render.custom_field | {"placeholder": "Batch ID"},
    )
    label = StringField(
        label="Label",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Name of consumable."},
        description="The label, i.e. the name of the consumable.",
    )
    consumable_type = SelectField(
        "Type",
        choices=[
            ("", "All"),
            ("antibody", "Antibody"),
            ("chemical", "Chemical"),
            ("enzyme", "Enzyme"),
        ],
        render_kw=render.select_field,
        description="""
        Batches can be added for all kinds of consumables like antibodies, chemicals,
        or enzmyes. This field allows  to select only batches of a certain type.
        """,
    )
    supplier = SelectField(
        "Supplier",
        choices=[("", "All")],
        render_kw=render.select_field,
        description="The supplier from which the batch was ordered.",
    )
    lot = StringField(
        "Lot",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Lot#"},
        description="""
        Batches have a field for lot numbers. This is usually only necessary 
        for batches of antibodies.
        """,
    )
    empty = SelectField(
        "Empty",
        choices=[
            ("all", "All"),
            ("empty", "Empty only"),
            ("not_empty", "Non-empty only"),
        ],
        default="all",
        render_kw=render.select_field,
        description="""
        Batches can be marked as empty once it was used up. However, 
        empty batches are kept in the database. This field allows to search 
        only for empty or non-empty batches.
        """,
    )
    in_use = SelectField(
        "In use",
        choices=[("all", "All"), ("in_use", "In use"), ("not_in_use", "Not in use")],
        default="all",
        render_kw=render.select_field,
    )
    order_by = SelectField(
        label="Order by",
        choices=[
            ("label", "Label"),
            ("id", "ID"),
            ("consumable_type", "Type"),
            ("supplier", "Supplier"),
            ("ordered", "Ordered"),
        ],
        default="label",
        render_kw=render.select_field,
        description="The column by which the results shall be ordered.",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        supplier = (
            Batch.query.with_entities(Batch.supplier, Batch.supplier)
            .order_by(Batch.supplier)
            .distinct()
        )
        self.supplier.choices += supplier

    def fields(self) -> list:
        return [
            self.id,
            self.consumable_type,
            self.label,
            self.supplier,
            self.lot,
            self.empty,
            self.in_use,
            self.order_by,
            self.ascending,
        ]


class EditBatch(BaseForm):
    """A form for adding or editing of a batch.

    Attributes
    ----------
    supplier : StringField
        The company/person that provided the batch. Limited to 64 characters.
    article_number : StringField
        The article number of the batch to easier find it again. Limited to
        32 characters.
    amount : StringField
        The amount (volume, weight, possibly also concentration) of the
        batch. Limited to 32 characters.
    date_ordered : DateField
        The date the batch was ordered.
    date_opened : DateField
        The date the batch was first used/opened.
    date_expiration : DateField
        The expiration date of the batch if provided.
    date_emptied : DateField
        The date the batch was emptied.
    price : DecimalField
        The price (if possible in €) of the batch.
    storage_place : StringField
        The place the batch is stored. Limited to 54 characters.
    lot : StringField
        The lot number of the batch.
    """

    supplier = StringField(
        "Supplier",
        validators=[DataRequired(), Length(max=64)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Supplier"},
    )
    article_number = StringField(
        "Article number",
        validators=[Optional(), Length(max=32)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Article number"},
    )
    amount = StringField(
        "Amount",
        validators=[Optional(), Length(max=32)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Amount"},
    )
    date_ordered = DateField(
        "Ordered",
        validators=[Optional()],
        render_kw=render.custom_field | {"type": "date"},
    )
    date_opened = DateField(
        "Opened",
        validators=[Optional()],
        render_kw=render.custom_field | {"type": "date"},
    )
    date_expiration = DateField(
        "Expiration date",
        validators=[Optional()],
        render_kw=render.custom_field | {"type": "date"},
    )
    date_emptied = DateField(
        "Emptied at",
        validators=[Optional()],
        render_kw=render.custom_field | {"type": "date"},
    )
    price = DecimalField(
        "Price",
        validators=[Optional(), NumberRange(min=0)],
        render_kw=render.custom_field
        | {"placeholder": "Price (€)", "min": 0, "type": "number", "step": 0.01},
    )
    storage_place = StringField(
        "Storage place",
        validators=[DataRequired(), Length(max=64)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Storage place"},
    )
    lot = StringField(
        "Lot number",
        validators=[DataRequired(), Length(max=64)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Lot number"},
    )



import io
from datetime import date

from flask import send_file
from pandas import DataFrame
from sqlalchemy import inspect

__all__ = ["Export"]


class Export:
    """A mixin for exporting instances to CSV, JSON, ...."""

    def to_dict(self) -> dict:
        """Create a dictionary from the object.

        Returns
        -------
        dict
            A dictionary. The keys are exactly labeled like the attributes of
            the instance.
        """

        inst = inspect(self).mapper.column_attrs
        return {c.key: getattr(self, c.key) for c in inst}

    @classmethod
    def to_df(cls, instances) -> DataFrame:
        return DataFrame(i.to_dict() for i in instances)

    @classmethod
    def _filename(cls) -> str:
        return cls.__name__ + "_" + date.today().isoformat()

    @classmethod
    def export_to_csv(cls, instances):
        with io.StringIO() as proxy:
            cls.to_df(instances).to_csv(proxy)
            mem = io.BytesIO(proxy.getvalue().encode("utf-8"))

        return send_file(
            mem,
            as_attachment=True,
            download_name=cls._filename() + ".csv",
            mimetype="text/csv",
        )

    @classmethod
    def export_to_json(cls, instances):
        with io.StringIO() as proxy:
            cls.to_df(instances).to_json(
                proxy, orient="records", date_format="iso", indent=2
            )
            mem = io.BytesIO(proxy.getvalue().encode("utf-8"))

        return send_file(
            mem,
            as_attachment=True,
            download_name=cls._filename() + ".json",
            mimetype="text/json",
        )

    @classmethod
    def to_pdf(cls, instances):
        mem = io.BytesIO()
        mem.seek(0)

        return send_file(
            mem,
            as_attachment=True,
            download_name=cls._filename() + ".pdf",
            mimetype="application/pdf",
        )



from labbase2.forms import render
from labbase2.forms.filters import make_lower, strip_input
from labbase2.forms.forms import EditEntityForm, FilterForm
from labbase2.models import Antibody
from sqlalchemy import func
from wtforms.fields import Field, IntegerField, SelectField, StringField
from wtforms.validators import DataRequired, Length, NumberRange, Optional

__all__ = ["FilterAntibodies", "EditAntibody"]


class FilterAntibodies(FilterForm):
    label = StringField(
        label="Label",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Label"},
        description="The label, i.e. the token of the entity.",
    )
    clone = StringField(
        label="Clone",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Clone"},
        description="Antibody clone if available.",
    )
    host = SelectField(
        label="Host",
        choices=[("", "All")],
        render_kw=render.select_field,
        description="The animal, in which the antibody was raised.",
    )
    antigen = StringField(
        label="Antigen",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Residue 46-89..."},
        description="The antigen that is recognized by the antibody.",
    )
    conjugate = SelectField(
        label="Conjugate",
        choices=[("all", "All"), ("any", "Any"), ("none", "None")],
        render_kw=render.select_field,
        description="""The conjugate of the antibody. Primary antibodies are 
        all antibodies without a conjugate.""",
    )
    order_by = SelectField(
        label="Order by",
        choices=[
            ("label", "Label"),
            ("id", "ID"),
            ("clone", "Clone"),
            ("host", "Host"),
            ("antigen", "Antigen"),
            ("conjugate", "Conjugate"),
        ],
        default="label",
        render_kw=render.select_field,
        description="The column by which the results shall be ordered.",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.conjugate.choices += (
            Antibody.query.with_entities(Antibody.conjugate, Antibody.conjugate)
            .order_by(func.lower(Antibody.conjugate))
            .filter(Antibody.conjugate.isnot(None))
            .distinct()
            .all()
        )

        self.host.choices += (
            Antibody.query.with_entities(Antibody.host, Antibody.host)
            .order_by(func.lower(Antibody.host))
            .distinct()
            .all()
        )

    def fields(self) -> list[Field]:
        return [
            self.id,
            self.label,
            self.clone,
            self.host,
            self.antigen,
            self.conjugate,
            self.ascending,
            self.order_by,
        ]


class EditAntibody(EditEntityForm):
    clone = StringField(
        label="Clone",
        validators=[Optional(), Length(max=32)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Clone"},
    )
    host = StringField(
        label="Host",
        validators=[DataRequired(), Length(max=64)],
        filters=[strip_input, make_lower],
        render_kw=render.custom_field | {"placeholder": "Host"},
    )
    antigen = StringField(
        "Antigen",
        validators=[DataRequired(), Length(max=64)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Antigen"},
    )
    specification = StringField(
        label="Specification",
        validators=[Optional(), Length(max=64)],
        filters=[strip_input, make_lower],
        render_kw=render.custom_field | {"placeholder": "Clonality"},
    )
    storage_temp = IntegerField(
        label="Storage Temperature",
        validators=[Optional(), NumberRange(min=-80, max=37)],
        render_kw=render.custom_field | {"min": -80, "max": 37, "type": "number"},
    )
    source = StringField(
        label="Source",
        validators=[Optional(), Length(max=64)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Source"},
    )
    conjugate = StringField(
        label="Conjugate",
        validators=[Optional(), Length(max=64)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Conjugate"},
    )



from flask import render_template
from flask_wtf import FlaskForm
from wtforms.fields import (
    BooleanField,
    Field,
    IntegerField,
    SelectField,
    StringField,
    SubmitField,
)
from wtforms.validators import NumberRange, Optional

from . import filters, render

__all__ = ["BaseForm", "FilterForm", "EditEntityForm"]


class BaseForm(FlaskForm):
    submit = SubmitField(label="Submit", render_kw=render.submit_field)

    def render(self, action: str = "", method: str = "GET") -> str:
        raise NotImplementedError


class FilterForm(BaseForm):
    id = IntegerField(
        label="ID",
        validators=[Optional(), NumberRange(min=1)],
        render_kw=render.custom_field | {"placeholder": "ID"},
        description="Internal database ID.",
    )
    ascending = BooleanField(
        label="Sort ascending",
        render_kw=render.boolean_field,
        default=True,
        description="Uncheck to sort results in descending order.",
    )
    order_by = SelectField(
        label="Order by",
        choices=[("id", "ID")],
        default="id",
        render_kw=render.select_field,
        description="The column by which the results shall be ordered.",
    )
    download_csv = SubmitField(label="Export to CSV", render_kw=render.submit_field)
    download_pdf = SubmitField(label="Export to PDF", render_kw=render.submit_field)
    download_excel = SubmitField(label="Export to Excel", render_kw=render.submit_field)

    def fields(self) -> list[Field]:
        raise NotImplementedError

    def render(self, action: str = "", method: str = "GET") -> str:
        return render_template("forms/filter.html", form=self, method=method)


class EditEntityForm(BaseForm):
    label = StringField(
        label="Label",
        validators=[Optional()],
        filters=[filters.strip_input],
        render_kw=render.custom_field | {"placeholder": "Name"},
        description="Must be unique among ALL database entries.",
    )



from flask_sqlalchemy import SQLAlchemy

# Is used in model classes, so has to be specified before model import.
db = SQLAlchemy()


from .antibody import Antibody, Dilution
from .base_entity import BaseEntity
from .chemical import Chemical, StockSolution
from .comment import Comment
from .consumable import Batch, Consumable
from .file import BaseFile, EntityFile
from .fly_stock import FlyStock, Modification
from .import_job import ColumnMapping, ImportJob
from .oligonucleotide import Oligonucleotide
from .plasmid import GlycerolStock, Plasmid, Preparation
from .request import Request
from .user import Permission, ResetPassword, User



from labbase2.forms import FilterForm, render
from labbase2.forms.filters import strip_input
from labbase2.models import User
from wtforms.fields import Field, SelectField, StringField, SubmitField
from wtforms.validators import Optional

__all__ = ["FilterOligonucleotide"]


class FilterOligonucleotide(FilterForm):
    label = StringField(
        label="Label",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Label"},
        description="The label, i.e. the name of the entity.",
    )
    owner_id = SelectField(
        label="Owner",
        choices=[(0, "All")],
        default=0,
        coerce=int,
        render_kw=render.select_field,
        description="The owner of the primer.",
    )
    sequence = StringField(
        label="Sequence",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "CAAAGCGGAGATAA..."},
        description="""
            The sequence or part of it. The search is case-insensitive. Underscores 
            can be used as wildcards. For instance, 'AC_T' finds all primers that 
            contain the motif 'ACNT' where 'N' can be any base.
            """,
    )
    description = StringField(
        label="Description",
        validators=[Optional()],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "forward m6 sequencing ..."},
        description="""
            A whitespace separated list of tags in the description of the primer. For
            instance, 'forward m6 sequencing' finds all primers with a description 
            that contains the words 'forward', 'm6', <strong>AND</strong> 'sequencing'.
            """,
    )
    download_fasta = SubmitField(label="Export to FASTA", render_kw=render.submit_field)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        user = (
            User.query.with_entities(User.id, User.username)
            .order_by(User.username)
            .all()
        )
        self.owner_id.choices += user
        self.order_by.choices += [
            ("label", "Label"),
            ("date_ordered", "Order date"),
            ("length", "Length"),
            ("sequence", "Sequence"),
            ("timestamp_edited", "Last edited"),
        ]

    def fields(self) -> list[Field]:
        return [
            self.id,
            self.label,
            self.owner_id,
            self.sequence,
            self.description,
            self.ascending,
            self.order_by,
        ]



from datetime import date, timedelta

from flask import Blueprint
from flask_login import current_user, login_required
from labbase2.models import GlycerolStock, Plasmid, Preparation, db
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required

from .forms import EditPreparation

__all__ = ["bp"]


# The blueprint to register all coming blueprints with.
bp = Blueprint(
    "preparations", __name__, url_prefix="/preparations", template_folder="templates"
)


@bp.route("/<int:plasmid_id>", methods=["POST"])
@login_required
@permission_required("Add preparations")
def add(plasmid_id: int):
    form = EditPreparation()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if Plasmid.query.get(plasmid_id) is None:
        return Message.ERROR(f"No plasmid with ID {plasmid_id}!")

    preparation = Preparation(owner_id=current_user.id, plasmid_id=plasmid_id)
    form.populate_obj(preparation)

    stock = GlycerolStock(
        plasmid_id=plasmid_id,
        owner_id=current_user.id,
        strain=form.strain.data,
        transformation_date=form.preparation_date.data - timedelta(days=1),
        disposal_date=form.preparation_date.data,
        storage_place="Immediately disposed.",
    )

    preparation.stock = stock

    try:
        db.session.add(preparation)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return Message.SUCCESS(
        f"Successfully added preparation to '{preparation.plasmid.label}'!"
    )


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
@permission_required("Add preparations")
def edit(id_: int):
    form = EditPreparation()

    if not form.validate():
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if (preparation := Preparation.query.get(id_)) is None:
        return Message.ERROR(f"No preparation with ID {id_}!")

    if preparation.owner_id != current_user.id:
        return Message.ERROR("Preparations can only be edited by owner!")

    form.populate_obj(preparation)

    try:
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return f"Successfully edited preparation!", 200


@bp.route("/<int:id_>", methods=["DELETE"])
@login_required
@permission_required("Add preparations")
def delete(id_: int):
    if not (preparation := Preparation.query.get(id_)):
        return Message.ERROR(f"No preparation with ID {id_}!")

    try:
        db.session.delete(preparation)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        return Message.ERROR(error)

    return Message.ERROR(f"Successfully deleted preparation {id_}!")



import io
from typing import Optional
from zipfile import ZIP_DEFLATED, ZipFile

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from flask import send_file
from flask_login import current_user
from labbase2.models import BaseEntity, db
from labbase2.models.fields import CustomDate
from labbase2.models.mixins import Filter, Sequence
from sqlalchemy import asc, desc

__all__ = ["Plasmid", "Preparation", "GlycerolStock"]


class Plasmid(BaseEntity, Sequence):
    """A class to represent plasmids in the database.

    Attributes
    ----------
    id : int
        The internal ID of this plasmis. The ID is unique among ALL entities
        in the datbase.
    insert : str
        The name of the insert. As explained for 'vector' a plasmid should be fully
        defined by the vector and the insert.
    vector : str
        The vector that was used for cloning. Usually, cloning means transferring an
        insert into a specific vector. By giving both the vector and the insert the
        plasmid should be fully defined.
    cloning_date : Date
        The date the plasmid was first created. This usually corresponds to the day
        of ligation.
    description : str
        A precise definition of the plasmid: What is it and what is it for. You may
        also give further information on how the plasmid was made.
    reference : str
        If the plasmid is published or was used somewhere it is possible to enter the
        DOI here.
    preparations : list[Preparation]
        A list of all preparations that were done of this plasmid.
    glycerol_stocks : list[GlycerolStock]
        A list of bacterial stocks that carry this plasmid.

    Notes
    -----
    As most other classes Plasmid inherits from Entity and therefor has three
    additional attributes: 'comments', 'files', and 'requests'. See 'Entity' for
    further explanations.
    """

    id = db.Column(db.Integer, db.ForeignKey("base_entity.id"), primary_key=True)
    insert = db.Column(db.String(128), nullable=False, info={"importable": True})
    vector = db.Column(db.String(256), info={"importable": True})
    cloning_date = db.Column(CustomDate, info={"importable": True})
    description = db.Column(db.String(2048), info={"importable": True})
    reference = db.Column(db.String(512), info={"importable": True})
    file_plasmid_id = db.Column(
        db.Integer, db.ForeignKey("base_file.id"), nullable=True
    )
    file_map_id = db.Column(db.Integer, db.ForeignKey("base_file.id"), nullable=True)

    file = db.relationship(
        "BaseFile",
        lazy=True,
        foreign_keys=[file_plasmid_id],
        single_parent=True,
        cascade="all, delete-orphan",
    )
    map = db.relationship(
        "BaseFile",
        lazy=True,
        foreign_keys=[file_map_id],
        single_parent=True,
        cascade="all, delete-orphan",
    )

    # One-to-many relationships.
    preparations = db.relationship(
        "Preparation",
        backref="plasmid",
        lazy=True,
        order_by="Preparation.emptied_date, Preparation.preparation_date",
    )
    glycerol_stocks = db.relationship(
        "GlycerolStock",
        backref="plasmid",
        lazy=True,
        order_by="GlycerolStock.disposal_date, GlycerolStock.transformation_date.desc()",
    )

    __mapper_args__ = {"polymorphic_identity": "plasmid"}

    def __len__(self):
        if record := self.seqrecord:
            return len(record)
        else:
            return 0

    @property
    def storage_place(self) -> Optional[str]:
        for preparation in self.preparations:
            if (
                preparation.date_emptied is not None
                and preparation.owner_id == current_user.id
            ):
                return preparation.restricted_storage_place

        return None

    @property
    def seqrecord(self) -> Optional[SeqRecord]:
        """The sequence of this plasmid.

        Returns
        -------
        Seq
            A biopython Seq object.

        Notes
        -----
        The sequence is read out from the plasmid filepath and consequently only
        available if such a filepath was uploaded.
        """

        if self.file_plasmid_id is None:
            return

        match self.file.path.suffix.lower():
            case ".gb" | ".gbk":
                format_ = "genbank"
            case ".dna":
                format_ = "snapgene"
            case ".xdna":
                format_ = "xdna"
            case _:
                return None

        try:
            record = SeqIO.read(self.file.path, format=format_)
        except Exception as error:
            print(error)
            return None
        else:
            return record

    @classmethod
    def to_zip(cls, entities):
        mem = io.BytesIO()

        with ZipFile(mem, "w", ZIP_DEFLATED, False) as archive:
            for plasmid in entities:
                if plasmid.file_plasmid_id:
                    archive.write(plasmid.file.path, plasmid.file.filename_exposed)
                if plasmid.file_map_id:
                    archive.write(plasmid.map.path, plasmid.map.filename_exposed)

        mem.seek(0)

        return send_file(
            mem,
            as_attachment=True,
            download_name="plasmids.zip",
        )


class Preparation(db.Model):
    """A specific preparation of a plasmid.

    Attributes
    ----------
    id : int
        The internal identifier of this preparation. This identifier is an integer
        that is not continuous with the identifiers of entities (antibodies, flies,
        plasmids, ...).
    plasmid_id : int
        The identifier of the plasmid that was prepared.
    owner_id : int
        The ID of the user who made this preparation.
    preparation_date : Date
        The date at which the preparation was done.
    method : str
        A short description of the method that was used for preparation. Most likely,
        this will be the name of the kit used.
    eluent : str
        The eluent. This will probably be either ddH2O or the elution buffer of the kit.
    concentration : float
        The concentration of this preparation. This should be in ng/ul or an
        appropriate concentration (ug/ml, ...).
    storage_place : str
        The location of this preparation.
    emptied_date : Date
        The date at which this preparation was used up.

    """

    id = db.Column(db.Integer, primary_key=True)
    plasmid_id = db.Column(db.Integer, db.ForeignKey("plasmid.id"), nullable=False)
    owner_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    preparation_date = db.Column(db.Date)
    method = db.Column(db.String(64))
    eluent = db.Column(db.String(32))
    concentration = db.Column(db.Integer())
    storage_place = db.Column(db.String(64))
    emptied_date = db.Column(db.Date)
    stock_id = db.Column(db.Integer, db.ForeignKey("glycerol_stock.id"), nullable=False)

    stock = db.relationship("GlycerolStock", backref="preparation", lazy=True)

    @property
    def restricted_storage_place(self) -> str:
        if current_user.id == self.owner_id:
            return self.storage_place
        else:
            return "Only accessible by owner of preparation!"


class GlycerolStock(db.Model, Filter):
    """A bacterial stock (glycerol stock) of a plasmid.

    Attributes
    ----------
    id : int
        The internal identifier of this bacterial stock. This identifier is an
        integer that is not continuous with the identifiers of entities (antibodies,
        flies, plasmids, ...).
    plasmid_id : int
        The internal identifier of the plasmid that was transformed into the
        bacteria.
    owner_id : int
        The ID of the person that transformed the plasmid into the bacteria. Maximum
        number of chars is 64. This might be changed in the future to an integer that
        refers to a person that is represented in the database.
    strain : str
        The token of the bacterial strain that was used for transformation,
        for instance 'DH10B'. Maximum number of chars is 64.
    transformation_date : Date
        The date the stock was created. This is the date of transformation.
    storage_place : str
        The storage place of the stock. This should include the -80°C freezer
        and the exact shelf in that freezer. Maximum number of chars is 128.
    disposal_date : Date
        The date the stock was disposed.
    """

    __tablename__: str = "glycerol_stock"

    id = db.Column(db.Integer, primary_key=True)
    plasmid_id = db.Column(db.Integer, db.ForeignKey("plasmid.id"), nullable=False)
    owner_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    strain = db.Column(db.String(64), nullable=False)
    transformation_date = db.Column(db.Date, nullable=False)
    storage_place = db.Column(db.String(128), nullable=False)
    disposal_date = db.Column(db.Date)

    @classmethod
    def filter_(cls, order_by: str = "label", ascending: bool = True, **fields):
        return super().filter_(order_by, ascending, **fields)

    @classmethod
    def _order_by(cls, order_by: str, ascending: bool) -> tuple:
        fnc = asc if ascending else desc
        if order_by == "label":
            field = Plasmid.label
        else:
            field = getattr(cls, order_by.strip())

        return (fnc(field),)

    @classmethod
    def _entities(cls) -> tuple:
        return cls, Plasmid.label

    @classmethod
    def _joins(cls) -> tuple:
        return (Plasmid,)



from labbase2.models import db, mixins
from labbase2.models.consumable import Consumable
from sqlalchemy import func

__all__ = ["Antibody", "Dilution"]


class Antibody(Consumable):
    """This class models a table for antibodies.

    Attributes
    ----------
    id : int
        An internal identifier of this antibody. This not to be set by the user but
        by the database.
    clone : str
        The clone from which this antibody was produced. Only relevant for monoclonal
        antibodies.
    host : str
        The host, i.e. die animal, in which this antibody has been raised. For
        nanobodies, something like 'Camelid' is appropriate.
    antigen : str
        The antigen against which this antibody has been raised.
    specification : str
        The clonality of the antibody. Usually one of 'Monoclonal', 'Polyclonal', or
        'Superclonal'.
    storage_temp : int
        The temperature in °C at which this antibody should be stored.
    source : str
        The source of this antibody. This is most important if the antibody is not
        commercially available.
    conjugate : str
        Some antibodies (e.g. secondaries) are conjugated with fluorophores or
        enzymes for detection.
    dilutions : list[Dilution]
        A list of dilutions. Each dilution indicates how this antibody should be
        diluted for a given application.
    """

    __tablename__: str = "antibody"

    id = db.Column(db.Integer, db.ForeignKey("consumable.id"), primary_key=True)
    clone = db.Column(db.String(32), info={"importable": True})
    host = db.Column(db.String(64), nullable=False, info={"importable": True})
    antigen = db.Column(db.String(256), nullable=False, info={"importable": True})
    specification = db.Column(db.String(64), info={"importable": True})
    storage_temp = db.Column(db.Integer, info={"importable": True})
    source = db.Column(db.String(64), info={"importable": True})
    conjugate = db.Column(db.String(64), info={"importable": True})

    # One-to-many relationships.
    dilutions = db.relationship(
        "Dilution",
        backref="antibody",
        lazy=True,
        order_by="Dilution.application, Dilution.timestamp_created.desc(), Dilution.id",
        cascade="all, delete-orphan",
    )

    __mapper_args__ = {"polymorphic_identity": "antibody"}

    def to_dict(self) -> dict:
        as_dict = super().to_dict()

        return as_dict | {"dilutions": [d.to_dict() for d in self.dilutions]}


class Dilution(db.Model, mixins.Export):
    """This table is considered to hold information about working dilutions
    for antibodies for different applications like immunostaining or western
    blot.

    Attributes
    ----------
    id : int
        An internal identifier of this dilution.
    antibody_id : int
        The ID of the antibody for the given entry.
    user_id : int
        ID of the user who determined this dilution (experimentally).
    application : str
        Specification of the application, e.g. 'Western blot' or 'Immunostaining'.
    dilution : str
        The dilution for the given antibody and application. This might be given in
        the form '1:x' or as an absolute concentration.
    reference : str
        A short explanation how this dilution was determined. This might refer to the
        spec sheet or the lab book.
    timestamp_created : DateTime
        The time at which this dilution was added to the database.
    timestamp_edited : DateTime
        The time at which this dilution was last edited. Might be `None` is it was
        never modified.
    """

    id = db.Column(db.Integer, primary_key=True)
    antibody_id = db.Column(db.Integer, db.ForeignKey("antibody.id"), nullable=False)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    application = db.Column(db.String(64), nullable=False)
    dilution = db.Column(db.String(32), nullable=False)
    reference = db.Column(db.String(2048), nullable=False, info={"importable": True})
    timestamp_created = db.Column(
        db.DateTime, server_default=func.now(timezone=True), info={"importable": True}
    )
    timestamp_edited = db.Column(
        db.DateTime(timezone=True),
        nullable=True,
        onupdate=func.now(timezone=True),
        info={"importable": True},
    )


# TODO: Implement an upvote system for dilutions.



from labbase2.forms import render
from labbase2.forms.filters import strip_input
from labbase2.forms.forms import BaseForm
from wtforms.fields import StringField, TextAreaField
from wtforms.validators import DataRequired, Length

__all__ = ["EditComment"]


class EditComment(BaseForm):
    """Form to edit a comment.

    Attributes
    ----------
    subject : StringField
        The subject of the comment.
    text : TextAreaField
        The actual comment. This is limited to 2048 characters.
    """

    subject = StringField(
        label="Subject",
        validators=[DataRequired(), Length(max=128)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Subject"},
    )
    text = TextAreaField(
        label="Comment",
        validators=[DataRequired(), Length(max=2048)],
        filters=[strip_input],
        render_kw=render.custom_field | {"placeholder": "Comment", "rows": 8},
    )



from datetime import date

from labbase2.models import db
from labbase2.models.base_entity import BaseEntity
from labbase2.models.fields import CustomDate
from labbase2.models.mixins.importer import Importer

__all__ = ["Modification", "FlyStock"]


class Modification(db.Model, Importer):
    """
    Information about modification of fly stocks. This should be things
    like re-establishing a dead fly stock. If the genotype is altered it
    might be rather better to create a new fly stock entry.

    Attributes
    ----------
    id : int
        An internal identifier for this modifcation.
    fly_id : int
        The identifier of the fly stock to which this modifcation belongs.
    user_id : int
        ID of the person that did the modification.
    date : date
        Date of the modification. Defaults to the current date.
    description : str
        A short and precise explanation of what was done.

    """

    id = db.Column(db.Integer, primary_key=True)
    fly_id = db.Column(db.Integer, db.ForeignKey("fly_stock.id"))
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"))
    date = db.Column(db.Date)
    description = db.Column(db.String(1024))


class FlyStock(BaseEntity):
    """A fly stock.

    Attributes
    ----------
    id : int
        The internal ID of this fly stock. This ID is unique among ALL entities
        in the database.
    short_genotype : str
        The complete genotype of this stock. Heterologeous chromosomes are
        separated by ';' and homologeous chromosomes are separated by '/'.
        The order of the chromosomes is 'x ; y ; 2 ; 3 ; 4'.
    chromosome_xa : str
    chromosome_xb : str
    chromosome_y : str
    chromosome_2a : str
    chromosome_2b : str
    chromosome_3a : str
    chromosome_3b : str
    chromosome_4a : str
    chromosome_4b : str
    location : str
        A string describing where this stock is located.
    created_date : date
        The date at which this stock was created.
    source : str
        The source of this stock. This might be either the token of the person
        that created this stock or Bloomington, VDRC, Kyoto, etc.
    documentation : str
        If this stock was created in our lab this should be a short
        description about how this stock was created or, even better,
        the pages in the lab book describing the generation of this stock.
    reference : str
        The DOI of a publication in which this stock was used.
    discarded_date : date
        The date at which this stock was discarded or died out.

    """

    id = db.Column(
        db.Integer,
        db.ForeignKey("base_entity.id"),
        primary_key=True,
        info={"importable": False},
    )
    short_genotype = db.Column(db.String(2048), info={"importable": True})
    chromosome_xa = db.Column(
        db.String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_xb = db.Column(
        db.String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_y = db.Column(
        db.String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_2a = db.Column(
        db.String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_2b = db.Column(
        db.String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_3a = db.Column(
        db.String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_3b = db.Column(
        db.String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_4a = db.Column(
        db.String(2048), nullable=False, default="+", info={"importable": True}
    )
    chromosome_4b = db.Column(
        db.String(2048), nullable=False, default="+", info={"importable": True}
    )
    location = db.Column(db.String(64), info={"importable": True})
    created_date = db.Column(CustomDate, info={"importable": True})
    source = db.Column(db.String(512), info={"importable": True})
    documentation = db.Column(db.String(2048), info={"importable": True})
    reference = db.Column(db.String(512), info={"importable": True})
    discarded_date = db.Column(CustomDate, info={"importable": True})

    # One-to-many relationships.
    modifications = db.relationship(
        "Modification",
        backref="fly_stock",
        order_by="Modification.date.desc()",
        lazy=True,
    )

    # Proper setup for joined table inheritance.
    __mapper_args__ = {"polymorphic_identity": "fly_stock"}

    @classmethod
    def _filters(cls, **fields) -> list:
        filters = []

        match fields.pop("discarded", "all"):
            case "discarded":
                filters.append(cls.discarded_date.isnot(None))
            case "recent":
                filters.append(cls.discarded_date.is_(None))
            case _:
                pass

        return super()._filters(**fields) + filters



import json
from typing import Iterable

import numpy as np
from numpy.typing import NDArray

__all__ = ["LCSFinder", "LCSResult"]


class LCSFinder:
    """A class to find matching primers for a target sequence.

    The class is initialized with the target sequence, which can be of any
    length. The instance can then be queried with any other sequence that
    consists only of "ATCG" and is not longer than 255 elements.

    Attributes
    ----------
    seq: NDArray
        The target sequence for which primers shall be found. The sequence is
        stored as a numpy array of dtype 'U1'.
    idx2len: NDArray
        A 2-dimensional array of unsigned 8-bit integers indicating the
        longest substring at each position. Rows correspond to positions in
        the queried sequence and columns correspond to positions in the
        sequence the Seeker class was initialized with. The number of rows is
        always 255, which therefor is the max length for the queried
        sequence. This allows it to use dtype uint8 to reduce the footprint.
    profile: NDArray
        A 1-dimensional array of dtype uint8 the same length as 'seq'. The
        profile indicates the longest substring at each position in 'seq'.
    """

    _bases: NDArray = np.array(["A", "T", "C", "G"], dtype=np.dtype("U1"))

    def __init__(self, seq: str):
        self.seq = np.array(list(seq.upper()), dtype=np.dtype("U1"))
        self.idx2len = np.zeros((255, len(self)), dtype=np.uint8)
        self.profile = np.zeros_like(self.seq, dtype=np.ubyte)
        self._base2idx = self._build_base2idx(self.seq)

    def __len__(self) -> int:
        return len(self.seq)

    def __call__(self, query: str) -> "LCSResult":
        """

        Parameters
        ----------
        query: Iterable
            The query sequence. Should be an iterable that yields single
            characters, e.g. a str or array of characters.

        Returns
        -------
        None
            This method doesn't return anything. It only sets attributes of
            the Seeker class, which allows it to extract relevant information
            afterwards.
        """

        if len(query) > 255:
            raise ValueError(f"Max supported query length is 255!")

        # Clear table from any previous searches.
        self.idx2len[...] = 0

        for i, base in enumerate(query.upper()):
            j = self._base2idx[base]
            self.idx2len[i, j] = self.idx2len[i - 1, j - 1] + 1

        return LCSResult(self.seq, self.idx2len)

    @classmethod
    def _build_base2idx(cls, seq) -> dict:
        """Build a dict of positions for each base in 'seq'.

        The dict is a helper for finding the longest common substring. For
        each of the four bases 'ACTG' the dict contains an NDArray with the
        positions of the base in 'seq'.

        Returns
        -------
        None
        """

        base2idx = {}

        for base in cls._bases:
            (idx,) = (seq == base).nonzero()  # 'nonzero' returns tuples.
            base2idx |= {base: idx}

        return base2idx


class LCSResult:

    def __init__(self, seq: NDArray, idx2len: NDArray):
        self.seq = seq
        self.profile = idx2len.max(axis=0)

        lcs_pos = self.profile.argmax()

        self.length = self.profile[lcs_pos]
        self.start = lcs_pos - self.length + 1
        self.lcs = seq[self.start : (self.start + self.length)]
        self.lcs = "".join(self.lcs)

    def to_jsarray(self) -> str:
        """Convert profile into a javascript array string.

        Returns
        -------
        str
            A string of the form '[1, 2, 3, ...]'.
        """

        return json.dumps(self.profile.tolist())



from wtforms.fields import Field
from wtforms.form import Form
from wtforms.validators import ValidationError

__all__ = [
    "ContainsSpecial",
    "ContainsNot",
    "ContainsNumber",
    "AllowCharacters",
    "ContainsLower",
    "ContainsUpper",
    "ContainsNotSpace",
    "AllASCII",
]


class RemoveCharacters:

    def __init__(self, chars: str):

        self.chars = chars

    def __call__(self, x) -> str:
        return "".join([c for c in x if c not in self.chars])


class AllowCharacters:

    def __init__(self, chars: str):
        self.chars = chars

    def __call__(self, form: Form, field: Field) -> None:
        data = field.data

        for char in data:
            if char not in self.chars:
                raise ValidationError(f"'{char}' is not a valid character.")


class ContainsNot:
    """A validator that checks that certain characters do not appear in the
    input.

    Attributes
    ----------
    forbidden: list[str]
        A list of strings that are not allowed in the input field.
    message: str, optional
        A message that is returned when validation fails.
    """

    def __init__(self, forbidden: list = None, message: str = None):
        self.forbidden = forbidden
        if not message:
            message = "Forbidden characters: " + ", ".join(self.forbidden)
        self.message = message

    def __call__(self, form: Form, field: Field) -> None:
        """Test if a field contains valid data.

        Parameters
        ----------
        form : Form
            The form, from which this validator is called.
        field : Field
            The field, of which the data shall be validated.

        Returns
        -------
        None

        Raises
        ------
        ValidationError if any forbidden character appears in the input field
        data.
        """

        data = field.data

        for char in self.forbidden:
            if char in data:
                raise ValidationError(self.message)


class ContainsLower:
    """ """

    def __init__(self, message: str = None):

        if not message:
            message = "Must contain at least one lowercase letter!"
        self.message = message

    def __call__(self, form: Form, field: Field) -> None:
        """Test if a field contains valid data.

        Parameters
        ----------
        form : Form
            The form, from which this validator is called.
        field : Field
            The field, of which the data shall be validated.

        Returns
        -------
        None

        Raises
        ------
        ValidationError if no character in the field data is lowercase.
        """

        data: str = field.data

        for char in data:
            if char.islower():
                return

        raise ValidationError(self.message)


class ContainsUpper:
    """ """

    def __init__(self, message: str = None):

        if not message:
            message = "Must contain at least one uppercase letter!"
        self.message = message

    def __call__(self, form: Form, field: Field) -> None:
        """Test if a field contains valid data.

        Parameters
        ----------
        form : Form
            The form, from which this validator is called.
        field : Field
            The field, of which the data shall be validated.

        Returns
        -------
        None

        Raises
        ------
        ValidationError if no character in the field data is uppercase.
        """

        data: str = field.data

        for char in data:
            if char.isupper():
                return

        raise ValidationError(self.message)


class ContainsNumber:
    """ """

    def __init__(self, message: str = None):

        if not message:
            message = "Must contain at least one number!"
        self.message = message

    def __call__(self, form: Form, field: Field) -> None:
        """Test if a field contains valid data.

        Parameters
        ----------
        form : Form
            The form, from which this validator is called.
        field : Field
            The field, of which the data shall be validated.

        Returns
        -------
        None

        Raises
        ------
        ValidationError if no character in the field data is a number.
        """

        data: str = field.data

        for char in data:
            if char.isdigit():
                return

        raise ValidationError(self.message)


class ContainsSpecial:
    """ """

    def __init__(self, message: str = None):

        if not message:
            message = "Must contain at least one special character!"
        self.message = message

    def __call__(self, form: Form, field: Field) -> None:
        """Test if a field contains valid data.

        Parameters
        ----------
        form : Form
            The form, from which this validator is called.
        field : Field
            The field, of which the data shall be validated.

        Returns
        -------
        None

        Raises
        ------
        ValidationError if no character in the field data is a number.
        """

        data: str = field.data

        for char in data:
            if not char.isalnum():
                return

        raise ValidationError(self.message)


class ContainsNotSpace:
    """ """

    def __init__(self, message: str = None):

        if not message:
            message = "Must not contain a whitespace character!"
        self.message = message

    def __call__(self, form: Form, field: Field) -> None:
        """Test if a field contains valid data.

        Parameters
        ----------
        form : Form
            The form, from which this validator is called.
        field : Field
            The field, of which the data shall be validated.

        Returns
        -------
        None

        Raises
        ------
        ValidationError if no character in the field data is a number.
        """

        data: str = field.data

        for char in data:
            if char.isspace():
                raise ValidationError(self.message)


class AllASCII:
    """ """

    def __init__(self, message: str = None):

        if not message:
            message = "Must contain ASCII characters only!"
        self.message = message

    def __call__(self, form: Form, field: Field) -> None:
        """Test if a field contains valid data.

        Parameters
        ----------
        form : Form
            The form, from which this validator is called.
        field : Field
            The field, of which the data shall be validated.

        Returns
        -------
        None

        Raises
        ------
        ValidationError if no character in the field data is a number.
        """

        data: str = field.data

        for char in data:
            if not char.isascii():
                raise ValidationError(self.message)



from flask import Blueprint
from flask import current_app as app
from flask_login import current_user, login_required
from labbase2.models import Dilution, db
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required

from .forms import EditDilution

__all__ = ["bp"]


bp = Blueprint("dilutions", __name__, url_prefix="/dilution", template_folder="templates")


@bp.route("/<int:antibody_id>", methods=["POST"])
@login_required
@permission_required("Add dilution")
def add(antibody_id: int):
    form = EditDilution()

    if not form.validate():
        app.logger.info(
            "Couldn't add dilution to antibody with ID %d due to invalid user input.", antibody_id
        )
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    dilution = Dilution(antibody_id=antibody_id, user_id=current_user.id)
    form.populate_obj(dilution)

    try:
        db.session.add(dilution)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        app.logger.warning(
            "Couldn't add dilution to database due to unknown database error: %s", error
        )
        return Message.ERROR(error)
    else:
        app.logger.info("Added new dilution with ID %d to database.", dilution.id)

    return Message.SUCCESS(f"Successfully added dilution to '{dilution.antibody.label}'!")


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
@permission_required("Add dilution")
def edit(id_: int):
    form = EditDilution()

    if not form.validate():
        app.logger.info("Couldn't edit dilution with ID %d due to invalid user input.", id_)
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if (dilution := Dilution.query.get(id_)) is None:
        app.logger.warning("Couldn't find dilution with ID %d.", id_)
        return Message.ERROR(f"No dilution found with ID {id_}!")

    if dilution.user_id != current_user.id:
        app.logger.warning(
            "Tried to edit dilution with ID %d that belongs to %s.", id_, dilution.user.username
        )
        return Message.ERROR(
            "Dilutions can only be edited by owner! Consider adding a new dilution instead."
        )

    form.populate_obj(dilution)

    try:
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        app.logger.warning(
            "Couldn't write changes for dilution with ID %d to database: %s", id_, error
        )
        return Message.ERROR(error)
    else:
        app.logger.info("Edited dilution with ID %d.", id_)

    return Message.SUCCESS(f"Successfully edited dilution '{dilution.id}'!")


@bp.route("<int:id_>", methods=["DELETE"])
@login_required
@permission_required("Add dilution")
def delete(id_: int):
    if (dilution := Dilution.query.get(id_)) is None:
        app.logger.warning("Couldn't find dilution with ID %d.", id_)
        return Message.ERROR(f"No dilution found with ID {id_}!")

    if dilution.user_id != current_user.id:
        app.logger.warning(
            "Tried to delete dilution with ID %d that belongs to %s.", id_, dilution.user.username
        )
        return Message.ERROR("Dilutions can only be deleted by owner!")

    try:
        db.session.delete(dilution)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        app.logger.warning(
            "Couldn't delete dilution with ID %d due to unknown database error.", id_
        )
        return Message.ERROR(error)
    else:
        app.logger.info("Deleted dilution with ID %d.", id_)

    return Message.ERROR(f"Successfully deleted dilution {id_}!")



from typing import Optional


__all__ = ["remove_whitespaces", "strip_input", "make_lower", "make_upper"]


def remove_whitespaces(x: Optional[str]) -> Optional[str]:
    """Remove whitespaces from a string

    Parameters
    ----------
    x : Optional[str], optional
        A string from which all whitespaces shall be removed or `None`. Defaults to
        `None`.

    Returns
    -------
    Optional[str]
        The original string `x` but with whitespaces removed or `None` if `x` was
        `None`.
    """

    if x is None:
        return None

    if not isinstance(x, str):
        raise ValueError("`x` has to be a string!")

    return "".join(x.split())


def strip_input(x: Optional[str] = None) -> Optional[str]:
    """A simple wrapper for the strip-method of strings applicable for use as a
    filter for form fields.

    Parameters
    ----------
    x : Optional[str], optional
        A string that is to be stripped or `None`. Defaults to `None`.

    Returns
    -------
    Optional[str]
        The original string `x` but stripped at both ends or `None` if `x` was `None`.
    """

    if x is None:
        return

    if not isinstance(x, str):
        raise ValueError("`x` has to be a string!")

    return x.strip()


def make_lower(x: Optional[str] = None) -> Optional[str]:
    """A simple wrapper for the lower-method of strings applicable for use as a
    filter for form fields.

    Parameters
    ----------
    x : Optional[str], optional
        A string that is to be stripped or `None`. Defaults to `None`.

    Returns
    -------
    Optional[str]
        The original string 'x' but with all characters being replaced by their
        lowercase equivalents or `None` if `x` was `None`.
    """

    if x is None:
        return None

    if not isinstance(x, str):
        raise ValueError("`x` has to be a string!")

    return x.lower()


def make_upper(x: Optional[str] = None) -> Optional[str]:
    """A simple wrapper for the upper-method of strings for use as a filter
    for form fields.

    Parameters
    ----------
    x : Optional[str], optional
        A string that is to be converted to uppercase or `None`. Defaults to `None`.

    Returns
    -------
    Optional[str]
        The original string 'x' but with all characters being replaced by their
        uppercase equivalents or `None` if `x` was `None`.
    """

    if x is None:
        return None

    if not isinstance(x, str):
        raise ValueError("`x` has to be a string!")

    return x.upper()



import datetime

from flask import Blueprint
from flask import current_app as app
from flask import flash, render_template, request
from flask_login import login_required
from labbase2.models import Batch, db
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required

from .forms import EditBatch, FilterBatch

__all__ = ["bp"]


# The blueprint to register all coming blueprints with.
bp = Blueprint("batches", __name__, url_prefix="/batch", template_folder="templates")


@bp.route("/", methods=["GET"])
@login_required
def index():
    page = request.args.get("page", 1, type=int)
    form = FilterBatch(request.args)

    data = form.data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = Batch.filter_(**data)
    except Exception as error:
        flash(str(error), "danger")
        app.logger.error("Couldn't filter batches: %s", error)
        entities = Batch.filter_(order_by="label")
    else:
        app.logger.debug("Found %d batches.", entities.count())

    return render_template(
        "batches/main.html",
        filter_form=form,
        add_form=EditBatch(formdata=None),
        entities=entities.paginate(page=page, per_page=app.config["PER_PAGE"]),
        total=Batch.query.count(),
        title="Batches",
    )


@bp.route("/<int:consumable_id>", methods=["POST"])
@login_required
@permission_required("Add consumable batches")
def add(consumable_id: int):
    form = EditBatch()

    if not form.validate():
        app.logger.info("Couldn't add batch due to invalid input.")
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    batch = Batch(consumable_id=consumable_id)
    form.populate_obj(batch)

    try:
        db.session.add(batch)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        app.logger.warning("Couldn't add batch due to unknown database error: %s", error)
        return Message.ERROR(error)
    else:
        app.logger.info("Added new batch with ID %5d.", batch.id)

    return Message.SUCCESS(f"Successfully added batch to '{batch.consumable.label}'!")


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
@permission_required("Add consumable batches")
def edit(id_: int):
    form = EditBatch()

    if not form.validate():
        app.logger.info("Couldn't edit batch with ID %d due to invalid user input.", id_)
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if (batch := Batch.query.get(id_)) is None:
        app.logger.warning("Couldn't find batch with ID %d.", id_)
        return Message.ERROR(f"No batch with ID {id_}!")

    form.populate_obj(batch)

    try:
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        app.logger.warning("Couldn't edit batch with ID %d due to unknown database error: %s", id_, error)
        return Message.ERROR(error)
    else:
        app.logger.info("Edited batch eith ID %d.", id_)

    return Message.SUCCESS(f"Successfully edited batch {id_}!")


@bp.route("/open/<int:id_>", methods=["PUT"])
@login_required
def in_use(id_: int):
    if (batch := Batch.query.get(id_)) is None:
        app.logger.warning("Couldn't find batch with ID %d.", id_)
        return Message.ERROR(f"No batch with ID {id_}!")

    if batch.date_opened:
        app.logger.warning("Batch with ID %d already marked open.", id_)
        return Message.WARNING(f"Batch {batch.id_} already marked open!")

    batch.date_opened = datetime.date.today()
    batch.in_use = True

    try:
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        app.logger.warning("Couldn't mark batch with ID %d as open: %s", id_, error)
        return Message.ERROR(error)
    else:
        app.logger.info("Marked batch with ID %d as open.", id_)

    return Message.SUCCESS(f"Marked batch {id_} as open!")


@bp.route("/empty/<int:id_>", methods=["PUT"])
@login_required
def emptied(id_: int):
    if (batch := Batch.query.get(id_)) is None:
        app.logger.warning("Couldn't find batch with ID %d.", id_)
        return Message.ERROR(f"No batch with ID {id_}!")

    if batch.date_emptied:
        app.logger.warning("Batch with ID %d already marked empty.", id_)
        return f"Batch {id_} already marked empty!", 200

    batch.date_emptied = datetime.date.today()
    batch.in_use = False

    try:
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        app.logger.warning("Couldn't mark batch with ID %d empty due to unknown database error: %s", id_, error)
        return Message.ERROR(error)
    else:
        app.logger.info("Marked batch with ID %d as empty.", id_)

    return Message.SUCCESS(f"Marked batch {id_} as empty!")


@bp.route("/<int:id_>", methods=["DELETE"])
@login_required
@permission_required("Add consumable batches")
def delete(id_: int):
    if (batch := Batch.query.get(id_)) is None:
        app.logger.warning("Couldn't find batch with ID %d.", id_)
        return Message.ERROR(f"No batch with ID {id_}!")

    try:
        db.session.delete(batch)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        app.logger.warning("Couldn't delete batch with ID %d due to unknown database error: %s", id_, error)
        return Message.ERROR(error)
    else:
        app.logger.info("Deleted batch with ID %d.", id_)

    return Message.SUCCESS(f"Successfully deleted batch {id_}!")


@bp.route("/<int:id_>/<string:format_>", methods=["GET"])
@login_required
def details(id_: int, format_: str):
    if (batch := Batch.query.get(id_)) is None:
        app.logger.warning("Couldn't find batch with ID %d.", id_)
        return Message.ERROR(f"No batch with ID {id_}!")

    match format_:
        case "long":
            template = "batches/details.html"
        case "tab":
            template = "batches/details-tab.html"
        case _:
            return Message.ERROR(f"Invalid format '{format_}'!")

    edit_form = EditBatch(None, obj=batch)

    return render_template(template, batch=batch, form=edit_form)



from http.client import responses

from flask import url_for
from flask_login import current_user, login_user
from sqlalchemy import select, func
from labbase2 import models
from labbase2.models import db


def test_login_get(app, client):
    with app.app_context(), client:
        url = url_for("auth.login")
        response = client.get(url)

    assert response.status_code == 200
    assert b"Sign in" in response.data


def test_login_with_wrong_user(app, client):
    with app.app_context(), client:
        url = url_for("auth.login")
        response = client.post(
            url,
            data={"email": "wrong_email@test.de", "password": "admin", "submit": True},
            follow_redirects=True,
        )

        assert response.status_code == 200
        assert b"Invalid email address or username!" in response.data
        assert current_user.is_anonymous
        assert not current_user.is_authenticated


def test_login_with_wrong_pw(app, client):
    with app.app_context(), client:
        url = url_for("auth.login")
        response = client.post(
            url,
            data={"email": "test@test.de", "password": "wrong_pw", "submit": True},
            follow_redirects=True,
        )

        assert response.status_code == 200
        assert b"Wrong password!" in response.data
        assert current_user.is_anonymous
        assert not current_user.is_authenticated


def test_login_inactive_user(app, client):
    with app.app_context(), client:
        user = db.session.get(models.User, 1)
        user.is_active = False

        db.session.commit()

        url = url_for("auth.login")
        response = client.post(
            url,
            data={"email": "test@test.de", "password": "admin", "submit": True},
            follow_redirects=True,
        )

        assert response.status_code == 200
        assert b"Your account is inactive!" in response.data
        assert current_user.is_anonymous
        assert not current_user.is_authenticated


def test_login_with_correct_data(app, client):
    with app.app_context(), client:
        url = url_for("auth.login")
        response = client.post(
            url,
            data={"email": "test@test.de", "password": "admin", "submit": True},
            follow_redirects=True,
        )

        assert response.status_code == 200
        assert b"Successfully logged in!" in response.data
        assert not current_user.is_anonymous
        assert current_user.is_authenticated
        assert current_user.username == "Max Mustermann"


def test_logout_user(app, client):
    with app.app_context(), app.test_request_context(), client:
        url = url_for("auth.logout")
        user = db.session.get(models.User, 1)
        login_user(user)
        response = client.get(url, follow_redirects=True)

        assert b"Successfully logged out!" in response.data
        assert current_user.is_anonymous
        assert not current_user.is_authenticated


def test_register_existing_email(app, client):
    with app.app_context(), app.test_request_context(), client:
        admin = db.session.get(models.User, 1)
        login_user(admin)

        url = url_for("auth.register")
        response = client.post(
            url,
            data={
                "first_name": "Maja",
                "last_name": "Musterfrau",
                "email": "test@test.de",
                "password": "This_isAPassword123",
                "password2": "This_isAPassword123",
                "submit": True,
            },
            follow_redirects=True,
        )

        user_count = db.session.scalar(select(func.count()).select_from(models.User))

        assert b"Email address already exists!" in response.data
        assert user_count == 1


def test_register_new_user(app, client):
    with app.app_context(), app.test_request_context(), client:
        admin = db.session.get(models.User, 1)
        login_user(admin)

        url = url_for("auth.register")
        response = client.post(
            url,
            data={
                "first_name": "Maja",
                "last_name": "Musterfrau",
                "email": "test2@test.de",
                "password": "This_isAPassword123",
                "password2": "This_isAPassword123",
                "submit": True,
            },
            follow_redirects=True,
        )

        user_count = db.session.scalar(select(func.count()).select_from(models.User))

        assert b"Successfully registered new user!" in response.data
        assert user_count == 2


def test_get_edit_user(app, client):
    with app.app_context(), app.test_request_context(), client:
        user = db.session.get(models.User, 1)
        login_user(user)
        url = url_for("auth.edit_user")
        response = client.get(url)

        assert response.status_code == 200
        assert b"Edit Profile" in response.data


def test_edit_user_with_wrong_pw(app, client):
    with app.app_context(), app.test_request_context(), client:
        user = db.session.get(models.User, 1)
        login_user(user)
        url = url_for("auth.edit_user")
        response = client.post(
            url,
            data={
                "first_name": "Maximilian",
                "last_name": "Mustermann",
                "email": "test@test.de",
                "timezone": "UTC",
                "password": "wrong_password",
            },
            follow_redirects=True,
        )

        assert response.status_code == 200
        assert b"Password is incorrect!" in response.data
        assert user.first_name == "Max"


def test_edit_user(app, client):
    with app.app_context(), app.test_request_context(), client:
        user = db.session.get(models.User, 1)
        login_user(user)
        url = url_for("auth.edit_user")
        response = client.post(
            url,
            data={
                "first_name": "Maximilian",
                "last_name": "Mustermann",
                "email": "test@test.de",
                "timezone": "UTC",
                "password": "admin",
            },
            follow_redirects=True,
        )

        assert response.status_code == 200
        assert b"Successfully edited user profile!" in response.data
        assert user.first_name == "Maximilian"


def test_remove_only_admin(app, client):
    with app.app_context(), app.test_request_context(), client:
        user = db.session.get(models.User, 1)
        login_user(user)
        url = url_for("auth.change_admin_status", id_=1)
        response = client.get(url, follow_redirects=True)

        assert response.status_code == 200
        assert b"Successfully changed admin setting for" in response.data
        assert b"No user with admin rights! Reverting previous change!" in response.data
        assert user.is_admin


def test_make_user_admin(app, client):
    with app.app_context(), app.test_request_context(), client:
        admin = db.session.get(models.User, 1)
        user = models.User(first_name="Maja", last_name="Musterfrau", email="test2@test.de")
        user.set_password("TestPassword")
        db.session.add(user)
        db.session.commit()

        login_user(admin)
        url = url_for("auth.change_admin_status", id_=2)
        response = client.get(url, follow_redirects=True)

        assert response.status_code == 200
        assert b"Successfully changed admin setting for" in response.data
        assert b"No user with admin rights! Reverting previous change!" not in response.data
        assert user.is_admin


def test_remove_admin_from_user(app, client):
    with app.app_context(), app.test_request_context(), client:
        admin = db.session.get(models.User, 1)
        user = models.User(first_name="Maja", last_name="Musterfrau", email="test2@test.de")
        user.is_admin = True
        user.set_password("TestPassword")
        db.session.add(user)
        db.session.commit()

        login_user(admin)
        url = url_for("auth.change_admin_status", id_=2)
        response = client.get(url, follow_redirects=True)

        assert response.status_code == 200
        assert b"Successfully changed admin setting for" in response.data
        assert b"No user with admin rights! Reverting previous change!" not in response.data
        assert not user.is_admin


# TODO Test auth.change_password
# TODO Test auth.change_permissions
# TODO Test auth.change_active_status


def test_inactivate_only_admin(app, client):
    with app.app_context(), app.test_request_context(), client:
        admin = db.session.get(models.User, 1)
        login_user(admin)

        assert admin.is_active

        url = url_for("auth.change_active_status", id_=1)
        response = client.get(url, follow_redirects=True)

        assert response.status_code == 200
        assert b"Successfully changed active setting for" in response.data
        assert b"No active user! Reverting previous change!" in response.data
        assert admin.is_active


def test_inactivate_user(app, client):
    with app.app_context(), app.test_request_context(), client:
        admin = db.session.get(models.User, 1)
        user = models.User(first_name="Maja", last_name="Musterfrau", email="test2@test.de")
        user.set_password("TestPassword")
        db.session.add(user)
        db.session.commit()

        login_user(admin)

        assert user.is_active

        url = url_for("auth.change_active_status", id_=2)
        response = client.get(url, follow_redirects=True)

        assert response.status_code == 200
        assert b"Successfully changed active setting for" in response.data
        assert not user.is_active


# TODO Test auth.create_password_reset


def test_users_get(app, client):
    with app.app_context(), app.test_request_context(), client:
        url = url_for("auth.users")

        admin = db.session.get(models.User, 1)
        login_user(admin)

        response = client.get(url)

        assert response.status_code == 200
        assert b"Mustermann" in response.data
        assert b"Musterfrau" not in response.data

        user = models.User(first_name="Maja", last_name="Musterfrau", email="test2@test.de")
        user.set_password("TestPassword")
        db.session.add(user)
        db.session.commit()

        response = client.get(url)

        assert response.status_code == 200
        assert b"Mustermann" in response.data
        assert b"Musterfrau" in response.data



from datetime import datetime, timedelta

from flask import current_app
from flask_login import current_user
from labbase2.models import db, mixins
from sqlalchemy import func, Table

__all__ = ["BaseEntity"]


base_to_base = db.Table(
    "base_to_base",
    db.Column("left_base_id", db.Integer, db.ForeignKey("base_entity.id"), primary_key=True),
    db.Column("right_base_id", db.Integer, db.ForeignKey("base_entity.id"), primary_key=True),
)


class BaseEntity(db.Model, mixins.Filter, mixins.Export, mixins.Importer):
    """The basic class from which every entry in the database inherits.

    Attributes
    ----------
    id : int
        The identifier of this entity. This ID is unique among all fly stocks,
        primer, antibodies, etc.
    label : str
        A string naming the entity. This must be unique among all entities and is
        supposed to be more human accessible than the ID.
    timestamp_created : DatetTime
        The time this entity was added to the database. This is not to be set by the
        user but will be set by the database.
    timestamp_edited : DatetTime
        The time this entity was last edited. Can be `None` if the entity was never
        modified. Update will automatically be triggered by the application and set
        by the database.
    owner_id : int
        The user ID of the person who imported this entity.
    origin : str
        A short string describing how the entity was added to the database.
    entity_type : str
        The type of the instance. This is needed for proper mapping in the database
        and setting up inheritance.
    comments : list[Comment]
        A list of all comments that were created for this entity.
    files : list[File]
        A list of all files that are attached to this entity. Note that some models
        feature additional files like plasmid maps for `Plasmid`. These file swill
        not be included in this list.
    requests : list[Request]
        A list of all requests that were made for this entity.

    Notes
    -----
    This base class allows that every child has a comments and files as well as a
    requests attribute that is stored in a single table in the final database. The
    inheritance is implemented as joined inheritance. That means that there is one
    table in the database for all entities with file_columns common to all children
    (those defined in this class) and one table for each child with file_columns
    specific to the respective class. Therefore, instances of child classes can only
    be constructed from the database by joining several tables.
    """

    __tablename__: str = "base_entity"

    id = db.Column(db.Integer, primary_key=True, info={"importable": False})
    label = db.Column(
        db.String(64),
        nullable=False,
        unique=True,
        index=True,
        info={"importable": True},
    )
    timestamp_created = db.Column(
        db.DateTime,
        nullable=False,
        server_default=func.now(),
        info={"importable": False},
    )
    timestamp_edited = db.Column(
        db.DateTime(), nullable=True, onupdate=func.now(), info={"importable": True}
    )
    owner_id = db.Column(
        db.Integer,
        db.ForeignKey("user.id"),
        nullable=False,
        default=lambda: current_user.id,
        info={"importable": True},
    )
    origin = db.Column(db.String(256), nullable=True, info={"importable": False})
    entity_type = db.Column(db.String(32), nullable=False, info={"importable": False})

    # One-to-many relationships.
    comments = db.relationship(
        "Comment",
        backref="entity",
        order_by="Comment.timestamp_created.desc()",
        lazy=True,
        cascade="all, delete-orphan",
    )
    files = db.relationship(
        "EntityFile",
        backref="entity",
        order_by="EntityFile.timestamp_uploaded.desc()",
        lazy=True,
        cascade="all, delete-orphan",
    )
    requests = db.relationship(
        "Request",
        backref="entity",
        order_by="Request.timestamp.desc()",
        lazy=True,
        cascade="all, delete-orphan",
    )

    # Self-referencing
    self_references = db.relationship(
        "BaseEntity",
        secondary=base_to_base,
        primaryjoin=id == base_to_base.c.left_base_id,
        secondaryjoin=id == base_to_base.c.right_base_id,
        backref="self_referenced_by",
    )

    # Proper setup for joined table inheritance.
    __mapper_args__ = {"polymorphic_identity": "base", "polymorphic_on": entity_type}

    __table_args__ = {"extend_existing": True}

    @property
    def deletable(self) -> bool:
        hours = current_app.config["DELETABLE_HOURS"]
        return (datetime.now() - self.timestamp_created) <= timedelta(hours=hours)

    def to_dict(self) -> dict:
        as_dict = super().to_dict()

        return as_dict | {
            "comments": [c.to_dict() for c in self.comments],
            "requests": [r.to_dict() for r in self.requests],
        }

    @classmethod
    def _filters(cls, **fields) -> list:
        filters = []

        owner_id = fields.pop("owner_id", 0)
        if owner_id != 0:
            filters.append(cls.owner_id == owner_id)

        return super()._filters(**fields) + filters



import secrets
from datetime import datetime, timedelta
from typing import Optional
from zoneinfo import ZoneInfo

from email_validator import validate_email
from flask import Blueprint, Response
from flask import current_app as app
from flask import flash, redirect, render_template, request, url_for
from flask_login import current_user, login_required, login_user, logout_user
from labbase2.models import BaseFile, Permission, ResetPassword, User, db
from labbase2.utils.permission_required import permission_required
from labbase2.views.files.routes import upload_file
from sqlalchemy import select, func
from sqlalchemy.exc import DataError

from .forms.edit import EditUserForm
from .forms.edit_permissions import EditPermissions
from .forms.login import LoginForm
from .forms.password import ChangePassword
from .forms.register import RegisterForm

__all__ = ["bp", "login"]


# The blueprint to register all coming blueprints with.
bp = Blueprint("auth", __name__, url_prefix="/auth", template_folder="templates")


# @bp.route("/", methods=["GET"])
# @login_required
# def index():
#     return render_template("user/index.html", user=current_user)


@bp.route("/login", methods=["GET", "POST"])
def login() -> str | Response:
    # Current user is already authenticated.
    if current_user.is_authenticated:
        flash("You are already logged in!", "warning")
        return redirect(url_for("base.index"))

    form = LoginForm()

    # POST fork with correct credentials.
    if form.validate_on_submit():
        user = db.session.scalar(select(User).where(User.email == form.email.data))

        if not user:
            app.logger.warning("Login attempt with invalid credentials")
            flash("Invalid email address or username!", "danger")
        elif not user.is_active:
            app.logger.info("Deactivate user tried to log in: %s", user.username)
            flash("Your account is inactive! Get in touch with an admin!", "warning")
        elif not user.verify_password(form.password.data):
            app.logger.warning("Login attempt with invalid password: %s", user.username)
            flash("Wrong password!", "danger")
        else:
            login_user(user, remember=form.remember_me.data)
            app.logger.info("Logged in: %s", user.username)
            flash("Successfully logged in!", "success")

            last_login = current_user.timestamp_last_login
            if last_login is None:
                flash("This is your first log in!", "info")
            else:
                tz = current_user.timezone
                formatted = last_login.astimezone(ZoneInfo(tz)).strftime("%B %d, %Y %I:%M %p")
                flash(f"Last login was on {formatted}!", "info")

            user.timestamp_last_login = func.now()
            try:
                db.session.commit()
            except Exception as error:
                db.error("Unable to save login time due to unknown database error: %s", error)

            next_page = request.args.get("next")

            if not next_page:  # or url_parse(next_page).netloc:
                next_page = url_for("base.index")

            return redirect(next_page)

    # GET fork or POST with wrong credentials.
    return render_template("auth/login.html", title="Login", login_form=form)


@bp.route("/logout", methods=["GET"])
@login_required
def logout() -> str | Response:
    app.logger.info("Logged out.")
    flash("Successfully logged out!", "success")

    logout_user()

    return redirect(url_for(".login"))


@bp.route("/register", methods=["GET", "POST"])
@login_required
@permission_required("Manage users")
def register():
    form = RegisterForm()

    # POST fork of view.
    if form.validate_on_submit():
        email = validate_email(form.email.data).normalized

        user_count_stm = select(func.count()).select_from(User).where(User.email == email)
        if db.session.scalar(user_count_stm) > 0:
            app.logger.warning("Tried to register new user with existing email address: %s", email)
            flash("Email address already exists!", "danger")
        else:
            user = User()
            form.populate_obj(user)
            user.email = email
            user.set_password(form.password.data)

            try:
                db.session.add(user)
                db.session.commit()
            except DataError as error:
                app.logger.warning("Could not write new user to database: %s", error)
                flash(str(error), "danger")
                db.session.rollback()
            else:
                app.logger.info("Registered new user: %s", user.username)
                flash("Successfully registered new user!", "success")

    elif form.submit():
        app.logger.warning("Could not register new user due to invalid input: %s", form.data)
        for field, errors in form.errors.items():
            for error in errors:
                flash(f"{field}: {error}", "danger")

    return render_template("auth/register.html", title="Register", form=form)


@bp.route("/edit", methods=["GET", "POST"])
@login_required
def edit_user():
    if request.method == "GET":
        form = EditUserForm(None, obj=current_user)
        return render_template("auth/edit-user.html", title="Edit profile", form=form)

    form = EditUserForm()

    if not form.validate():
        app.logger.warning("Could not edit user due to invalid input.")
        for error in form.errors:
            flash(str(error), "danger")
        return render_template("auth/edit-user.html", title="Edit profile", form=form)

    if not current_user.verify_password(form.password.data):
        app.logger.warning("Entered incorrect password.")
        flash("Password is incorrect!", "danger")
    else:
        form.populate_obj(current_user)

        if form.file.data:
            file = upload_file(form, BaseFile)
            current_user.picture = file
            file.resize(512)

        try:
            db.session.commit()
        except Exception as error:
            app.logger.warning("Could not update user: %s", error)
            flash(str(error), "error")
            db.session.rollback()
        else:
            app.logger.info("Updated user profile.")
            flash("Successfully edited user profile!", "success")

    return render_template("auth/edit-user.html", title="Edit profile", form=form)


@bp.route("/change-password", methods=["GET", "POST"], defaults={"key": None})
@bp.route("/change-password/<string:key>", methods=["GET", "POST"])
@login_required
def change_password(key: Optional[str] = None):
    if key and (reset := db.session.get(ResetPassword, key)) is not None:
        if datetime.now() > reset.timeout:
            flash("Token to reset password has expired!")
            db.session.delete(reset)
        else:
            user = reset.user
            logout_user()
            login_user(user)

    form = ChangePassword()

    if form.validate_on_submit():

        if reset is not None:
            reset.user.set_password(form.new_password.data)
            reset.user.resets.clear()
            updated = True

        elif current_user.verify_password(form.old_password.data):
            current_user.set_password(form.new_password.data)
            updated = True

        else:
            flash("Old password incorrect!", "danger")
            updated = False

        if updated:
            try:
                db.session.commit()
            except Exception as error:
                db.session.rollback()
                flash(str(error), "danger")
            else:
                flash("Your password has been updated!", "success")

    else:
        for field, errors in form.errors.items():
            for error in errors:
                flash(": ".join([field, error]), "danger")

    return render_template("auth/change-password.html", form=form)


@bp.route("/change-permissions", methods=["GET"], defaults={"id_": None})
@bp.route("/change-permissions/<int:id_>", methods=["GET", "POST"])
@login_required
@permission_required("Manage users")
def change_permissions(id_: int = None):
    users = db.session.scalars(
        select(User).where(User.is_active).order_by(User.last_name, User.first_name)
    )

    if id_ is None:
        user = current_user
    else:
        user = db.session.get(User, id_)

    form = EditPermissions(**user.form_permissions)

    if form.validate_on_submit():
        user.permissions.clear()
        for field in form:
            if not field.type == "BooleanField":
                continue
            if not field.data:
                continue

            permission = db.session.get(Permission, field.name.replace("_", " ").capitalize())
            user.permissions.append(permission)

        db.session.commit()
        app.logger.info("Updated permissions for user: %s", user.username)
        flash(f"Successfully updated permissions for {user.username}!", "success")

    return render_template(
        "auth/edit-permissions.html",
        form=form,
        selected_user=user,
        users=users,
    )


@bp.route("/change-admin-status/<int:id_>", methods=["GET"])
@login_required
@permission_required("Manage users")
def change_admin_status(id_: int):
    user = db.session.get(User, id_)

    if user is None:
        flash(f"No user with ID {id_})!", "danger")
    else:
        user.is_admin = not user.is_admin

        try:
            db.session.commit()
        except Exception as error:
            db.session.rollback()
            flash(str(error), "danger")
        else:
            flash(f"Successfully changed admin setting for {user.username}!", "success")

    user_count_stm = select(func.count()).select_from(User).where(User.is_admin)
    if db.session.scalar(user_count_stm) == 0:
        flash("No user with admin rights! Reverting previous change!", "warning")
        user.is_admin = True
        db.session.commit()

    if request.referrer is None:
        return redirect(url_for(".users"))
    else:
        return redirect(request.referrer)


@bp.route("/change-active-status/<int:id_>", methods=["GET"])
@login_required
@permission_required("Manage users")
def change_active_status(id_: int):
    user = db.session.get(User, id_)

    if user is None:
        flash(f"No user with ID {id_})!", "danger")
    else:
        user.is_active = not user.is_active

        try:
            db.session.commit()
        except Exception as error:
            db.session.rollback()
            flash(str(error), "danger")
        else:
            flash(f"Successfully changed active setting for {user.username}!", "success")

    user_count_stm = select(func.count()).select_from(User).where(User.is_active)
    if db.session.scalar(user_count_stm) == 0:
        flash("No active user! Reverting previous change!", "warning")
        user.is_active = True
        db.session.commit()

    if request.referrer is None:
        return redirect(url_for(".users"))
    else:
        return redirect(request.referrer)


@bp.route("/reset-password/<int:id_>", methods=["GET"])
@login_required
@permission_required("Manage users")
def create_password_reset(id_: int):
    user = db.session.get(User, id_)

    if user is None:
        flash(f"No user with ID {id_})!", "danger")
        return redirect(request.referrer)

    user.resets.clear()

    db.session.commit()

    expires_after = app.config.get("RESET_EXPIRES", 10)
    reset = ResetPassword(
        token=secrets.token_urlsafe(32),
        user_id=id_,
        timeout=datetime.now() + timedelta(minutes=expires_after),
    )

    db.session.add(reset)
    db.session.commit()

    flash(f"Generated a password reset . Expires: {reset.timeout.isoformat()}!", "success")
    flash(f"Visit: {url_for('auth.change_password', key=reset.token)}")

    if request.referrer is None:
        return redirect(url_for(".users"))
    else:
        return redirect(request.referrer)


@bp.route("/users", methods=["GET"])
@login_required
def users():
    users = db.session.scalars(
        select(User).order_by(
            User.is_active.desc(), User.is_admin.desc(), User.last_name, User.first_name
        )
    )
    return render_template("auth/users.html", users=users, title="Users")



from typing import Optional

from labbase2.utils.message import Message

__all__ = ["errors2messages"]


def errors2messages(field_errors: Optional[dict[str, str]]) -> list[str]:
    """Turns field errors of forms into a list of error messages.

    Parameters
    ----------
    field_errors : Optional[dict[str, str]]
        A dictionary of errors as produced by the `errors` property of forms. If the
        form was not validated yet, this property will be `None`, in which case this
        function will return an empty list.

    Returns
    -------
    list[str]
        A list of HTML-formatted error messages produced by successive calls to
        `Message.ERROR`. A list entry is created for each error in each form field.
        If the form was not validated yet, the list will be empty.
    """

    messages: list[str] = []

    if field_errors is None:
        return messages

    for field, errors in field_errors.items():
        for error in errors:
            messages.append(Message.ERROR(f"{field}: {error}"))

    return messages



from sqlite3 import IntegrityError

from flask import Blueprint
from flask import current_app as app
from flask import flash, render_template, request
from flask_login import current_user, login_required
from labbase2.models import Antibody, db
from labbase2.utils.message import Message
from labbase2.utils.permission_required import permission_required
from labbase2.views.batches.forms import EditBatch
from labbase2.views.comments.forms import EditComment
from labbase2.views.files.forms import UploadFile
from labbase2.views.requests.forms import EditRequest

from . import dilutions
from .dilutions.forms import EditDilution
from .forms import EditAntibody, FilterAntibodies


__all__ = ["bp"]


bp = Blueprint("antibodies", __name__, url_prefix="/antibody", template_folder="templates")

bp.register_blueprint(dilutions.bp)


@bp.route("/", methods=["GET"])
@login_required
def index():
    page = request.args.get("page", 1, type=int)
    form = FilterAntibodies(request.args)

    data = form.data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = Antibody.filter_(**data)
    except Exception as error:
        flash(str(error), "danger")
        app.logger.error("Couldn't filter antibodies: %s", error)
        entities = Antibody.filter_(order_by="label")
    else:
        app.logger.debug("Found %d antibodies.", entities.count())

    return render_template(
        "antibodies/main.html",
        filter_form=form,
        import_file_form=UploadFile(),
        add_form=EditAntibody(formdata=None),
        entities=entities.paginate(page=page, per_page=app.config["PER_PAGE"]),
        total=Antibody.query.count(),
        title="Antibodies",
    )


@bp.route("/<int:id_>", methods=["GET"])
@login_required
def details(id_: int):
    if (antibody := Antibody.query.get(id_)) is None:
        app.logger.warning("Couldn't find antibody with ID %s.", id_)
        return Message.ERROR(f"No antibody found with ID {id_}!")

    return render_template(
        "antibodies/details.html",
        antibody=antibody,
        form=EditAntibody(formdata=None, obj=antibody),
        comment_form=EditComment,
        request_form=EditRequest,
        file_form=UploadFile,
        batch_form=EditBatch,
        dilution_form=EditDilution,
    )


@bp.route("/", methods=["POST"])
@login_required
@permission_required("Add antibodies")
def add():
    form = EditAntibody()

    if not form.validate():
        app.logger.info("Couldn't add antibody due to invalid input.")
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    antibody = Antibody()
    antibody.origin = f"Created via import form by {current_user.username}."
    form.populate_obj(antibody)

    try:
        db.session.add(antibody)
        db.session.commit()
    except IntegrityError as error:
        db.session.rollback()
        app.logger.info("Couldn't add antibody due to integrity error.")
        return Message.ERROR(error)
    except Exception as error:
        db.session.rollback()
        app.logger.warning(
            "Couldn't add antibody to database due to unknown database error: %s", error
        )
        return Message.ERROR(error)
    else:
        app.logger.info("Added new antibody with ID %5d.", antibody.id)

    return Message.SUCCESS(f"Successfully added antibody '{antibody.label}'!")


@bp.route("/<int:id_>", methods=["PUT"])
@login_required
@permission_required("Add antibodies")
def edit(id_: int):
    form = EditAntibody()

    if not form.validate():
        app.logger.info("Couldn't edit antibody with ID %d due to invalid user input.", id_)
        return "<br>".join(Message.ERROR(error) for error in form.errors)

    if not (antibody := Antibody.query.get(id_)):
        app.logger.warning("Couldn't find antibody with ID %d.", id_)
        return Message.ERROR(f"No antibody found with ID {id_}!")

    form.populate_obj(antibody)

    try:
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        app.logger.warning("Couldn't edit antibody with ID %d due to unknown database error.", id_)
        return Message.ERROR(error)
    else:
        app.logger.info("Edited antibody with ID %d.", id_)

    return Message.SUCCESS(f"Successfully edited antibody '{antibody.label}'!")


@bp.route("/delete/<int:id_>", methods=["DELETE"])
@login_required
@permission_required("Add antibodies")
def delete(id_):
    if (antibody := Antibody.query.get(id_)) is None:
        app.logger.warning("Couldn't find antibody with ID %d.", id_)
        return Message.ERROR(f"No antibody found with ID {id_}!")

    try:
        db.session.delete(antibody)
        db.session.commit()
    except Exception as error:
        db.session.rollback()
        app.logger.warning(
            "Couldn't delete antibody with ID %d due to unknown database error.", id_
        )
        return Message.ERROR(error)
    else:
        app.logger.info("Deleted antibody with ID %d.", id_)

    return Message.SUCCESS(f"Successfully deleted antibody '{antibody.label}'!")


@bp.route("/export/<string:format_>/", methods=["GET"])
@login_required
@permission_required("Export content")
def export(format_: str):
    data = FilterAntibodies(request.args).data
    del data["submit"]
    del data["csrf_token"]

    try:
        entities = Antibody.filter_(**data)
    except Exception as error:
        app.logger.warning("Couldn't export antibodies: %s.", error)
        return Message.ERROR(error)

    match format_:
        case "csv":
            return Antibody.export_to_csv(entities)
        case "json":
            return Antibody.export_to_json(entities)
        case _:
            app.logger.warning("Tried to export antibodies with unsupported format: %s.", format_)
            return Message.ERROR(f"Unsupported format '{format_}'!")



from flask import Blueprint, render_template
from flask_login import login_required

__all__ = ["bp"]


# The blueprint to register all coming blueprints with.
bp = Blueprint("base", __name__, template_folder="templates")


@bp.route("/", methods=["GET"])
@login_required
def index() -> str:
    return render_template("base/index.html", title="Home")


# TODO: Add changelog.
